from Elasticipy.FourthOrderTensor import StiffnessTensor

C = StiffnessTensor.from_MP("mp-1048")
C.save_to_txt("TiNi.txt")
C.save_to_txt("TiNi.txt", matrix_only=True)

C2 = StiffnessTensor.from_txt_file("TiNi.txt")
Cu2= StiffnessTensor.from_txt_file("TiNi.txt")





import numpy as np



class JohnsonCook:
    def __init__(self, A, B, n, C=None, eps_dot_ref=1.0, m=None, T0=25, Tm=None):
        """
        Constructor for a Jonhson-Cook (JC) model.

        The JC model is an exponential-law strain hardening model, which can take into account strain-rate sensibility
        and temperature-dependence (although they are not mandatory). See notes for details.

        Parameters
        ----------
        A : float
            Yield stress
        B : float
            Work hardening coefficient
        n : float
            Work hardening exponent
        C : float, optional
            Strain-rate sensitivity coefficient
        eps_dot_ref : float, optional
            Reference strain-rate
        m : float, optional
            Temperature sensitivity exponent
        T0 : float, optional
            Reference temperature
        Tm : float, optional
            Melting temperature (at which the flow stress is zero)

        Notes
        -----
        The flow stress (:math:`\\sigma`) depends on the strain (:math:`\\varepsilon`),
        the strain rate :math:`\\dot{\\varepsilon}` and
        the temperature (:math:`T`) so that:

        .. math::

                \\sigma = \\left(A + B\\varepsilon^n\\right)
                        \\left(1 + C\\log\\left(\\frac{\\varepsilon}{\\dot{\\varepsilon}_0}\\right)\\right)
                        \\left(1-\\theta^m\\right)

        with

        .. math::

                \\theta = \\begin{cases}
                            \\frac{T-T_0}{T_m-T_0} & \\text{if } T<T_m\\\\
                            1                      & \\text{otherwise}
                            \\end{cases}
        """
        self.A = A
        self.B = B
        self.C = C
        self.n = n
        self.m = m
        self.eps_dot_ref = eps_dot_ref
        self.T0 = T0
        self.Tm = Tm

    def flow_stress(self, eps_p, eps_dot=None, T=None):
        """
        Compute the flow stress from the Johnson-Cook model

        Parameters
        ----------
        eps_p : float or list or tuple or numpy.ndarray
            Equivalent plastic strain
        eps_dot : float or list or tuple or numpy.ndarray, optional
            Equivalent plastic strain rate. If float, the strain-rate is supposed to be homogeneous for every value of
            eps_p.
        T : float or list or tuple or np.ndarray
            Temperature. If float, the temperature is supposed to be homogeneous for every value of eps_p.
        Returns
        -------
        numpy.ndarray
            Flow stress
        """
        eps_p = np.asarray(eps_p)
        stress = (self.A + self.B * eps_p**self.n)

        if eps_dot is not None:
            eps_dot = np.asarray(eps_dot)
            if (self.C is None) or (self.eps_dot_ref is None):
                raise ValueError('C and eps_dot_ref must be defined for using a rate-dependent model')
            stress *= (1 + self.C * np.log(eps_dot / self.eps_dot_ref))

        if T is not None:
            T = np.asarray(T)
            if self.T0 is None or self.Tm is None or self.m is None:
                raise ValueError('T0, Tm and m must be defined for using a temperature-dependent model')
            theta = (T - self.T0) / (self.Tm - self.T0)
            theta = np.clip(theta, None, 1.0)
            stress *= (1 - theta**self.m)

        return stress


    def compute_strain(self, stress, T=None):
        """
        Given the equivalent stress, compute the strain

        Parameters
        ----------
        stress : float or numpy.ndarray
            Equivalent stress tom compute the stress from
        T : float or list or tuple or numpy.ndarray
            Temperature

        Returns
        -------
        numpy.ndarray
            Equivalent strain
        """
        stress = np.asarray(stress)
        if T is None:
            theta_m=0.0
        else:
            if self.T0 is None or self.Tm is None or self.m is None:
                raise ValueError('T0, Tm and m must be defined for using a temperature-dependent model')
            else:
                theta = (T - self.T0) / (self.Tm - self.T0)
                theta_m = np.clip(theta**self.m, None, 1.0)
        return (1/self.B * ( stress / (1 - theta_m) - self.A))**(1/self.n)



import sys

import numpy as np
from PyQt5.QtWidgets import (
    QApplication, QMainWindow, QComboBox, QGridLayout, QLabel,
    QLineEdit, QPushButton, QVBoxLayout, QHBoxLayout, QWidget, QFrame, QMessageBox
)
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from Elasticipy.CrystalSymmetries import SYMMETRIES
from Elasticipy.FourthOrderTensor import StiffnessTensor

WHICH_OPTIONS = {'Mean': 'mean', 'Max': 'max', 'Min': 'min', 'Std. dev.': 'std'}

class ElasticityGUI(QMainWindow):
    def __init__(self):
        super().__init__()
        self.coefficient_fields = {}
        self.setWindowTitle("Elasticipy - GUI")
        self.initUI()

    def selected_symmetry(self):
        symmetry_name = self.symmetry_selector.currentText()
        symmetry = SYMMETRIES[symmetry_name]
        if symmetry_name == "Trigonal" or symmetry_name == "Tetragonal":
            space_group_text = self.point_group_selector.currentText()
            try:
                symmetry = symmetry[space_group_text]
            except KeyError:
                # If no SG is selected, just choose one
                symmetry = list(symmetry.values())[0]
        elif symmetry_name == "Monoclinic":
            diad_index = self.diag_selector.currentText()
            symmetry = symmetry[diad_index]
        return symmetry

    def initUI(self):
        # Main layout
        main_layout = QVBoxLayout()

        #######################################################################################
        # Material's symmetry and other parameters
        #######################################################################################
        selectors_layout = QHBoxLayout()

        # Symmetry selection
        self.symmetry_selector = QComboBox()
        self.symmetry_selector.addItems(SYMMETRIES.keys())
        self.symmetry_selector.currentIndexChanged.connect(self.update_fields)
        selectors_layout.addWidget(QLabel("Crystal symmetry:"))
        selectors_layout.addWidget(self.symmetry_selector)

        # Space Group selection
        self.point_group_selector = QComboBox()
        self.point_group_selector.addItems(['', ''])
        self.point_group_selector.currentIndexChanged.connect(self.update_fields)
        selectors_layout.addWidget(QLabel("Point group:"))
        selectors_layout.addWidget(self.point_group_selector)

        # Diad selection
        self.diag_selector = QComboBox()
        self.diag_selector.addItems(SYMMETRIES['Monoclinic'].keys())
        self.diag_selector.currentIndexChanged.connect(self.update_fields)
        selectors_layout.addWidget(QLabel("Diad convention:"))
        selectors_layout.addWidget(self.diag_selector)

        # Add horizontal separator
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)

        # Add selectors_layout to main layout
        main_layout.addLayout(selectors_layout)
        main_layout.addWidget(separator)

        #######################################################################################
        # Matrix components
        #######################################################################################
        grid = QGridLayout()
        for i in range(6):
            for j in range(i, 6):
                field = QLineEdit()
                field.setPlaceholderText(f"C{i+1}{j+1}")
                self.coefficient_fields[(i, j)] = field
                field.textChanged.connect(self.update_dependent_fields)
                grid.addWidget(field, i, j)
        main_layout.addLayout(grid)

        #######################################################################################
        # Bottom panel
        #######################################################################################
        bottom_layout = QHBoxLayout()

        ############################################
        # Plotting options
        ############################################
        left_panel_layout = QVBoxLayout()

        # E, G or nu selector
        parameter_layout = QHBoxLayout()
        self.plotting_selector = QComboBox()
        self.plotting_selector.addItems(['Young modulus', 'Shear modulus', 'Poisson ratio', 'Linear compressibility'])
        self.plotting_selector.currentIndexChanged.connect(self.update_plotting_selectors)
        parameter_layout.addWidget(QLabel("Parameter:"))
        parameter_layout.addWidget(self.plotting_selector)
        left_panel_layout.addLayout(parameter_layout)

        # Plotting style
        style_layout = QHBoxLayout()
        self.plot_style_selector = QComboBox()
        self.plot_style_selector.addItems(['3D', 'Sections', 'Pole Figure'])
        self.plot_style_selector.currentIndexChanged.connect(self.update_plotting_selectors)
        style_layout.addWidget(QLabel("Plot type:"))
        style_layout.addWidget(self.plot_style_selector)
        left_panel_layout.addLayout(style_layout)

        # 'which' selector
        which_layout = QHBoxLayout()
        self.which_selector = QComboBox()
        self.which_selector.addItems(WHICH_OPTIONS.keys())
        which_layout.addWidget(QLabel("Value:"))
        which_layout.addWidget(self.which_selector)
        left_panel_layout.addLayout(which_layout)

        # Plot button
        self.calculate_button = QPushButton("Plot")
        self.calculate_button.clicked.connect(self.calculate_and_plot)
        left_panel_layout.addWidget(self.calculate_button)

        # Fill space
        left_panel_layout.addStretch()
        bottom_layout.addLayout(left_panel_layout)

        ############################################
        # Plotting area
        ############################################
        # Add separator between the top and the bottom panels
        separator2 = QFrame()
        separator2.setFrameShape(QFrame.HLine)
        separator2.setFrameShadow(QFrame.Sunken)
        main_layout.addWidget(separator2)

        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        bottom_layout.addWidget(self.canvas)

        #######################################################################################
        # Main widget
        #######################################################################################
        main_layout.addLayout(bottom_layout)
        central_widget = QWidget()
        central_widget.setLayout(main_layout)
        self.setCentralWidget(central_widget)

        #######################################################################################
        # Initialize at load
        #######################################################################################
        self.symmetry_selector.setCurrentText('Triclinic')
        self.plotting_selector.setCurrentText('Young modulus')
        self.which_selector.setEnabled(False)

    def update_fields(self):
        # Deactivate unused fields
        active_fields = self.selected_symmetry().required
        for (i, j), field in self.coefficient_fields.items():
            if (i, j) in active_fields:
                field.setEnabled(True)
                if field.text() == "0":
                    field.setText('')
            else:
                field.setEnabled(False)
                field.setText('0')

        # Turn on/off SG selection
        selected_symmetry_name = self.symmetry_selector.currentText()
        trig_or_tetra = selected_symmetry_name in ("Trigonal", "Tetragonal")
        self.point_group_selector.setEnabled(trig_or_tetra) # Turn on/off SG selector
        if trig_or_tetra:
            # Change list of possible SGs
            self.point_group_selector.setEnabled(True)
            space_groups = SYMMETRIES[selected_symmetry_name].keys()
            for i in range(len(space_groups)):
                self.point_group_selector.setItemText(i, list(space_groups)[i])
        self.diag_selector.setEnabled(selected_symmetry_name == "Monoclinic")

    def update_plotting_selectors(self):
        if (self.plotting_selector.currentText() == "Young modulus" or
            self.plotting_selector.currentText() == "Linear compressibility" or
                self.plot_style_selector.currentIndex() == 1):
            self.which_selector.setEnabled(False)
        else:
            self.which_selector.setEnabled(True)

    def calculate_and_plot(self):
        """Collect entries and compute the stiffness tensor"""
        coefficients = np.zeros((6, 6))
        for (i, j), field in self.coefficient_fields.items():
            try:
                coefficients[i, j] = float(field.text())
            except ValueError:
                coefficients[i, j] = 0
        C = np.array(coefficients)
        Csym = C + np.tril(C.T, -1) # Rebuild the lower triangular part

        try:
            stiff = StiffnessTensor(Csym)
            self.figure.clear()
            requested_value = self.plotting_selector.currentText()
            if requested_value == "Young modulus":
                value = stiff.Young_modulus
                plot_kwargs = {}
            elif requested_value == 'Linear compressibility':
                value = stiff.linear_compressibility
                plot_kwargs = {}
            else:
                if requested_value == 'Shear modulus':
                    value = stiff.shear_modulus
                else:
                    value = stiff.Poisson_ratio
                plot_kwargs = {'which': WHICH_OPTIONS[self.which_selector.currentText()]}
            if self.plot_style_selector.currentIndex() == 0:
                value.plot3D(fig=self.figure, **plot_kwargs)
            elif self.plot_style_selector.currentIndex() == 1:
                value.plot_xyz_sections(fig=self.figure)
            else:
                value.plot_as_pole_figure(fig=self.figure, **plot_kwargs)
            self.canvas.draw()

        except ValueError as inst:
            QMessageBox.critical(self, "Singular stiffness", inst.__str__(), QMessageBox.Ok)


    def update_dependent_fields(self):
        symmetry = self.selected_symmetry()
        for equality in symmetry.equal:
            try:
                ref_value = float(self.coefficient_fields[equality[0]].text())
                for index in equality[1]:
                        self.coefficient_fields[index].setText(f"{ref_value}")
            except ValueError:
                pass
        for opposite in symmetry.opposite:
            try:
                ref_value = float(self.coefficient_fields[opposite[0]].text())
                for index in opposite[1]:
                        self.coefficient_fields[index].setText(f"{-ref_value}")
            except ValueError:
                pass
        if symmetry.C11_C12:
            try:
                C11 = float(self.coefficient_fields[(0, 0)].text())
                C12 = float(self.coefficient_fields[(0, 1)].text())
                for index in symmetry.C11_C12:
                    self.coefficient_fields[index].setText(f"{0.5*(C11-C12)}")
            except ValueError:
                pass

def crystal_elastic_plotter():
    app = QApplication(sys.argv)
    window = ElasticityGUI()
    window.show()
    sys.exit(app.exec_())



import numpy as np
from matplotlib import transforms as mtransforms
from matplotlib import scale as mscale
from matplotlib.ticker import FixedLocator
from matplotlib.projections.polar import ThetaFormatter

def format_coord(phi, theta):
    """
    Format the coordinates as follows: 'phi=val, theta=val', wheres vals are in degrees
    :param phi: azimuth
    :param theta: colatitude
    """
    return ('azimuth=%0.1f\N{DEGREE SIGN}, '
            'colatitude=%0.1f\N{DEGREE SIGN}'
            % (np.degrees(phi) % 360, np.degrees(theta)))
    
class StereographicScale(mscale.ScaleBase):
    name = 'stereographic'

    def set_default_locators_and_formatters(self, axis):
        axis.set_major_locator(FixedLocator(np.radians(np.arange(0, 90, 20))))
        axis.set_major_formatter(ThetaFormatter())  # Use the same formatter as for phi (called theta in polar)
        axis.set_minor_formatter(ThetaFormatter())

    def get_transform(self):
        return self.StereographicTransform()

    def limit_range_for_scale(self, vmin, vmax, minpos):
        return 0, min(vmax, np.pi / 2)

    class StereographicTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = True

        def transform_non_affine(self, theta):
            return np.tan(theta / 2)

        def inverted(self):
            return StereographicScale.InvertedStereographicTransform()

    class InvertedStereographicTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = True

        def transform_non_affine(self, r):
            return 2 * np.arctan(r)

        def inverted(self):
            return StereographicScale.StereographicTransform()


mscale.register_scale(StereographicScale)


class LambertScale(mscale.ScaleBase):
    name = 'lambert'

    def set_default_locators_and_formatters(self, axis):
        axis.set_major_locator(FixedLocator(np.radians(np.arange(0, 90, 15))))
        axis.set_major_formatter(ThetaFormatter())  # Use the same formatter as for phi (called theta in polar)
        axis.set_minor_formatter(ThetaFormatter())

    def get_transform(self):
        return self.LambertTransform()

    def limit_range_for_scale(self, vmin, vmax, minpos):
        return 0, min(vmax, np.pi / 2)

    class LambertTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = True

        def transform_non_affine(self, theta):
            return 2 * np.sin(theta / 2)

        def inverted(self):
            return LambertScale.InvertedLambertTransform()

    class InvertedLambertTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = True

        def transform_non_affine(self, r):
            theta = np.zeros(len(r))
            theta[r >= np.sqrt(2)] = np.pi / 2
            theta[r < np.sqrt(2)] = 2 * np.arcsin(r[r < np.sqrt(2)] / 2)
            return theta

        def inverted(self):
            return LambertScale.LambertTransform()


mscale.register_scale(LambertScale)

def add_polefigure(fig, *args, projection='stereographic', **kwargs):
    if projection.lower() == 'equal area':
        projection = 'lambert'
    ax = fig.add_subplot(*args, projection='polar', **kwargs)
    ax.set_rlim(0.0, np.pi / 2)
    ax.set_ylim(0.0, 2*np.pi)
    ax.set_rscale(projection)
    ax.format_coord = format_coord
    return ax


from Elasticipy.FourthOrderTensor import StiffnessTensor
from scipy.spatial.transform import Rotation
import matplotlib as mpl
mpl.use('Qt5Agg')   # Ensure interactive plot

# First, let consider the NiTi material:
C = StiffnessTensor.fromCrystalSymmetry(symmetry='monoclinic', diad='y', phase_name='TiNi',
                                        C11=231, C12=127, C13=104,
                                        C22=240, C23=131, C33=175,
                                        C44=81, C55=11, C66=85,
                                        C15=-18, C25=1, C35=-3, C46=3)
print(C)

# Let's have a look on its Young modulus
E = C.Young_modulus
# See min/max values
print(E)
# Now illustrate the spatial dependence
E.plot_xyz_sections()   # As 2D sections...
E.plot3D()                # ...or in 3D
E.plot_as_pole_figure()     # or even with PF
print(E.max())

# Apply a random rotation on stiffness tensor
rotation = Rotation.from_euler('zxz', [0, 45, 0], degrees=True)
Crot = C*rotation
# Check that the Young modulus has changed as well
Crot.Young_modulus.plot3D()

# Now let's consider the shear modulus
G = C.shear_modulus
G.plot_xyz_sections()   # Plot sections with min, max and mean
G.plot3D()     # And plot it in 3D
print(G.min())
print(G.max())

# Finally, let's have a look on the Poisson ratio
nu = C.Poisson_ratio
nu.plot_xyz_sections()
nu.plot3D(which='max')
print(nu.min())
print(nu.max())

# Now let consider a finite set of orientations
oris = Rotation.random(1000)
Crotated = C*oris  # Compute the Voigt average
Cvoigt = Crotated.Voigt_average()
print(Cvoigt.Young_modulus) # Look at the corresponding Young modulis
print(C.Voigt_average().Young_modulus) # Compare with infinite number of orientations





import numpy as np
import re

from Elasticipy.SecondOrderTensor import SymmetricSecondOrderTensor, rotation_to_matrix, is_orix_rotation
from Elasticipy.StressStrainTensors import StrainTensor, StressTensor
from Elasticipy.SphericalFunction import SphericalFunction, HyperSphericalFunction
from scipy.spatial.transform import Rotation
from Elasticipy.CrystalSymmetries import SYMMETRIES
from copy import deepcopy

def _parse_tensor_components(prefix, **kwargs):
    pattern = r'^{}(\d{{2}})$'.format(prefix)
    value = dict()
    for k, v in kwargs.items():
        match = re.match(pattern, k)  # Extract 'C11' to '11' and so
        if match:
            value[match.group(1)] = v
    return value


def _indices2str(ij):
    return f'{ij[0] + 1}{ij[1] + 1}'


def voigt_indices(i, j):
    """
    Translate the two-index notation to one-index notation

    Parameters
    ----------
    i : int or np.ndarray
        First index
    j : int or np.ndarray
        Second index

    Returns
    -------
    Index in the vector of length 6
    """
    voigt_mat = np.array([[0, 5, 4],
                          [5, 1, 3],
                          [4, 3, 2]])
    return voigt_mat[i, j]


def unvoigt_index(i):
    """
    Translate the one-index notation to two-index notation

    Parameters
    ----------
    i : int or np.ndarray
        Index to translate
    """
    inverse_voigt_mat = np.array([[0, 0],
                                  [1, 1],
                                  [2, 2],
                                  [1, 2],
                                  [0, 2],
                                  [0, 1]])
    return inverse_voigt_mat[i]


def _compute_unit_strain_along_direction(S, m, n, direction='longitudinal'):
    if not isinstance(S, ComplianceTensor):
        S = S.inv()
    if direction == 'transverse':
        ein_str = 'ijkl,pi,pj,pk,pl->p'
        return np.einsum(ein_str, S.full_tensor(), m, m, n, n)
    elif direction =='longitudinal':
        ein_str = 'ijkl,pi,pk,pj,pl->p'
        return np.einsum(ein_str, S.full_tensor(), m, m, n, n)
    else:
        ein_str = 'ijkk,pi,pj->p'
        return np.einsum(ein_str, S.full_tensor(), m, m)


def _isotropic_matrix(C11, C12, C44):
    return np.array([[C11, C12, C12, 0, 0, 0],
                     [C12, C11, C12, 0, 0, 0],
                     [C12, C12, C11, 0, 0, 0],
                     [0, 0, 0, C44, 0, 0],
                     [0, 0, 0, 0, C44, 0],
                     [0, 0, 0, 0, 0, C44]])

def _check_definite_positive(mat):
    try:
        np.linalg.cholesky(mat)
    except np.linalg.LinAlgError:
        eigen_val = np.linalg.eigvals(mat)
        raise ValueError('The input matrix is not definite positive (eigenvalues: {})'.format(eigen_val))


def _is_single_rotation(rotation):
    if isinstance(rotation, Rotation):
        return rotation.single
    elif is_orix_rotation(rotation):
        return rotation.size == 1
    else:
        raise TypeError('The input argument must be of class scipy.transform.Rotation or '
                        'orix.quaternion.rotation.Rotation')

def _rotate_tensor(full_tensor, rotation):
    rot_mat = rotation_to_matrix(rotation)
    str_ein =  '...im,...jn,...ko,...lp,mnop->...ijkl'
    return np.einsum(str_ein, rot_mat, rot_mat, rot_mat, rot_mat, full_tensor)

class SymmetricTensor:
    """
    Template class for manipulating symmetric fourth-order tensors.

    Attributes
    ----------
    matrix : np.ndarray
        (6,6) matrix gathering all the components of the tensor, using the Voigt notation.
    symmetry : str
        Symmetry of the tensor

    """
    tensor_name = 'Symmetric'
    voigt_map = np.ones((6, 6))
    C11_C12_factor = 0.5
    C46_C56_factor = 1.0
    component_prefix = 'C'

    def __init__(self, M, phase_name=None, symmetry='Triclinic', orientations=None,
                 check_symmetry=True, check_positive_definite=False):
        """
        Construct of stiffness tensor from a (6,6) matrix.

        The input matrix must be symmetric, otherwise an error is thrown (except if check_symmetry==False, see below)

        Parameters
        ----------
        M : np.ndarray
            (6,6) matrix corresponding to the stiffness tensor, written using the Voigt notation, or array of shape
            (3,3,3,3).
        phase_name : str, default None
            Name to display
        symmetry : str, default Triclinic
            Name of the crystal's symmetry
        check_symmetry : bool, optional
            Whether to check or not that the input matrix is symmetric.
        check_positive_definite : bool, optional
            Whether to check or not that the input matrix is definite positive
        """
        M = np.asarray(M)
        if M.shape == (6,6):
            matrix = M
        elif M.shape == (3,3,3,3):
            matrix = self._full_to_matrix(M)
        else:
            raise ValueError('The input matrix must of shape (6,6)')
        if check_symmetry and not np.all(np.isclose(matrix, matrix.T)):
            raise ValueError('The input matrix must be symmetric')
        if check_positive_definite:
            _check_definite_positive(matrix)

        self.matrix = matrix
        self.phase_name = phase_name
        self.symmetry = symmetry
        self.orientations = orientations

        for i in range(0,6):
            for j in range(0,6):
                def getter(obj, I=i, J=j):
                    return obj.matrix[I, J]
                getter.__doc__ = f"Returns the ({i + 1},{j + 1}) component of the {self.tensor_name} matrix."
                component_name = 'C{}{}'.format(i+1, j+1)
                setattr(self.__class__, component_name, property(getter))  # Dynamically create the property

    def __repr__(self):
        if self.phase_name is None:
            heading = '{} tensor (in Voigt notation):\n'.format(self.tensor_name)
        else:
            heading = '{} tensor (in Voigt notation) for {}:\n'.format(self.tensor_name, self.phase_name)
        print_symmetry = '\nSymmetry: {}'.format(self.symmetry)
        msg = heading + self.matrix.__str__() + print_symmetry
        if self.orientations is not None:
            msg = msg + '\n{} orientations'.format(len(self))
        return msg

    def __len__(self):
        if self.orientations is None:
            return 1
        else:
            return len(self.orientations)

    def full_tensor(self):
        """
        Returns the full (unvoigted) tensor, as a [3, 3, 3, 3] array

        Returns
        -------
        np.ndarray
            Full tensor (4-index notation)
        """
        i, j, k, ell = np.indices((3, 3, 3, 3))
        ij = voigt_indices(i, j)
        kl = voigt_indices(k, ell)
        m = self.matrix[ij, kl] / self.voigt_map[ij, kl]
        if self.orientations is None:
            return m
        else:
            return _rotate_tensor(m, self.orientations)

    @classmethod
    def _full_to_matrix(cls, full_tensor):
        ij, kl = np.indices((6, 6))
        i, j = unvoigt_index(ij).T
        k, ell = unvoigt_index(kl).T
        return full_tensor[i, j, k, ell] * cls.voigt_map[ij, kl]

    def rotate(self, rotation):
        """
        Apply a single rotation to a tensor, and return its component into the rotated frame.

        Parameters
        ----------
        rotation : Rotation or orix.quaternion.rotation.Rotation
            Rotation to apply

        Returns
        -------
        SymmetricTensor
            Rotated tensor
        """
        if _is_single_rotation(rotation):
            rotated_tensor = _rotate_tensor(self.full_tensor(), rotation)
            rotated_matrix = self._full_to_matrix(rotated_tensor)
            return self.__class__(rotated_matrix)
        else:
            raise ValueError('The rotation to apply must be single')

    def _unrotate(self):
        unrotated_tensor = deepcopy(self)
        unrotated_tensor.orientations = None
        return unrotated_tensor

    def __add__(self, other):
        if isinstance(other, np.ndarray):
            if other.shape == (6, 6):
                mat = self.matrix + other
            elif other.shape == (3, 3, 3, 3):
                mat = self._full_to_matrix(self.full_tensor() + other)
            else:
                raise ValueError('The input argument must be either a 6x6 matrix or a (3,3,3,3) array.')
        elif isinstance(other, SymmetricTensor):
            if type(other) == type(self):
                mat = self.matrix + other.matrix
            else:
                raise ValueError('The two tensors to add must be of the same class.')
        else:
            raise ValueError('I don''t know how to add {} with {}.'.format(type(self), type(other)))
        return self.__class__(mat)

    def __sub__(self, other):
        if isinstance(other, SymmetricTensor):
            return self.__add__(-other.matrix)
        else:
            return self.__add__(-other)

    def __mul__(self, other):
        if isinstance(other, SymmetricSecondOrderTensor):
            return SymmetricSecondOrderTensor(self * other.matrix)
        elif isinstance(other, np.ndarray):
            if other.shape[-2:] == (3, 3):
                if self.orientations is None:
                    return np.einsum('ijkl,...kl->...ij', self.full_tensor(), other)
                else:
                    return np.einsum('qijkl,...kl->q...ij', self.full_tensor(), other)
        elif isinstance(other, Rotation) or is_orix_rotation(other):
            if _is_single_rotation(other):
                return self.rotate(other)
            else:
                return self.__class__(self.matrix, symmetry=self.symmetry, orientations=other, phase_name=self.phase_name)
        else:
            return self.__class__(self.matrix * other, symmetry=self.symmetry)

    def __rmul__(self, other):
        if isinstance(other, (Rotation, float, int, np.number)) or is_orix_rotation(other):
            return self * other
        else:
            raise NotImplementedError('A fourth order tensor can be left-multiplied by rotations or scalar only.')

    def __truediv__(self, other):
        if isinstance(other, (float, int, np.number)):
            return self.__class__(self.matrix / other, symmetry=self.symmetry)
        else:
            raise NotImplementedError

    def __eq__(self, other):
        if isinstance(other, SymmetricTensor):
            return np.all(self.matrix == other.matrix) and np.all(self.orientations == other.orientations)
        elif isinstance(other, np.ndarray) and other.shape == (6,6):
            return np.all(self.matrix == other)
        else:
            raise NotImplementedError('The element to compare with must be a fourth-order tensor '
                                      'or an array of shape (6,6).')


    def _orientation_average(self):
        mean_full_tensor = np.mean(self.full_tensor(), axis=0)
        mean_matrix = self._full_to_matrix(mean_full_tensor)
        return self.__class__(mean_matrix)

    @classmethod
    def _matrixFromCrystalSymmetry(cls, symmetry='Triclinic', point_group=None, diad='y', prefix=None, **kwargs):
        if prefix is None:
            prefix = cls.component_prefix
        values = _parse_tensor_components(prefix, **kwargs)
        C = np.zeros((6, 6))
        symmetry = symmetry.capitalize()
        if ((symmetry == 'tetragonal') or (symmetry == 'trigonal')) and (point_group is None):
            raise ValueError('For tetragonal and trigonal symmetries, the point group is mandatory.')
        tetra_1 = ['4', '-4', '4/m']
        tetra_2 = ['4mm', '-42m', '422', '4/mmm']
        trigo_1 = ['3', '-3']
        trigo_2 = ['32', '-3m', '3m']
        if point_group is not None:
            if (point_group in tetra_1) or (point_group in tetra_2):
                symmetry = 'Tetragonal'
            elif (point_group in trigo_1) or (point_group in trigo_2):
                symmetry = 'Trigonal'
        symmetry_description = SYMMETRIES[symmetry]
        if symmetry == 'Tetragonal':
            if point_group in tetra_1:
                symmetry_description = symmetry_description[', '.join(tetra_1)]
            else:
                symmetry_description = symmetry_description[', '.join(tetra_2)]
        elif symmetry == 'Trigonal':
            if point_group in trigo_1:
                symmetry_description = symmetry_description[', '.join(trigo_1)]
            else:
                symmetry_description = symmetry_description[', '.join(trigo_2)]
        elif symmetry == 'Monoclinic':
            symmetry_description = symmetry_description["Diad || " + diad]
        for required_field in symmetry_description.required:
            C[required_field] = values[_indices2str(required_field)]

        # Now apply relationships between components
        for equality in symmetry_description.equal:
            for index in equality[1]:
                C[index] = C[equality[0]]
        for opposite in symmetry_description.opposite:
            for index in opposite[1]:
                C[index] = -C[opposite[0]]
        C11_C12 = symmetry_description.C11_C12
        if C11_C12:
            for index in C11_C12:
                C[index] = (C[0, 0] - C[0, 1]) * cls.C11_C12_factor

        if symmetry == 'Trigonal':
            C[3, 5] = cls.C46_C56_factor * C[3, 5]
            C[4, 5] = cls.C46_C56_factor * C[4, 5]

        return C + np.tril(C.T, -1)

    @classmethod
    def fromCrystalSymmetry(cls, symmetry='Triclinic', point_group=None, diad='y', phase_name=None, prefix=None,
                            **kwargs):
        """
        Create a fourth-order tensor from limited number of components, taking advantage of crystallographic symmetries

        Parameters
        ----------
        symmetry : str, default Triclinic
            Name of the crystallographic symmetry
        point_group : str
            Point group of the considered crystal. Only used (and mandatory) for tetragonal and trigonal symmetries.
        diad : str {'x', 'y'}, default 'x'
            Alignment convention. Sets whether x||a or y||b. Only used for monoclinic symmetry.
        phase_name : str, default None
            Name to use when printing the tensor
        prefix : str, default None
            Define the prefix to use when providing the components. By default, it is 'C' for stiffness tensors, 'S' for
            compliance.
        kwargs
            Keywords describing all the necessary components, depending on the crystal's symmetry and the type of tensor.
            For Stiffness, they should be named as 'Cij' (e.g. C11=..., C12=...).
            For Comliance, they should be named as 'Sij' (e.g. S11=..., S12=...).
            See examples below. The behaviour can be overriten with the prefix option (see above)

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        StiffnessTensor.isotropic : creates an isotropic stiffness tensor from two paremeters (e.g. E and v).

        Notes
        -----
        The relationships between the tensor's components depend on the crystallogrpahic symmetry [1]_.

        References
        ----------
        .. [1] Nye, J. F. Physical Properties of Crystals. London: Oxford University Press, 1959.

        Examples
        --------
        >>> from Elasticipy.FourthOrderTensor import StiffnessTensor\n
        >>> StiffnessTensor.fromCrystalSymmetry(symmetry='monoclinic', diad='y', phase_name='TiNi',
        ...                                     C11=231, C12=127, C13=104,
        ...                                     C22=240, C23=131, C33=175,
        ...                                     C44=81, C55=11, C66=85,
        ...                                     C15=-18, C25=1, C35=-3, C46=3)
        Stiffness tensor (in Voigt notation) for TiNi:
        [[231. 127. 104.   0. -18.   0.]
         [127. 240. 131.   0.   1.   0.]
         [104. 131. 175.   0.  -3.   0.]
         [  0.   0.   0.  81.   0.   3.]
         [-18.   1.  -3.   0.  11.   0.]
         [  0.   0.   0.   3.   0.  85.]]
        Symmetry: monoclinic

        >>> from Elasticipy.FourthOrderTensor import ComplianceTensor\n
        >>> ComplianceTensor.fromCrystalSymmetry(symmetry='monoclinic', diad='y', phase_name='TiNi',
        ...                                      S11=8, S12=-3, S13=-2,
        ...                                      S22=8, S23=-5, S33=10,
        ...                                      S44=12, S55=116, S66=12,
        ...                                      S15=14, S25=-8, S35=0, S46=0)
        Compliance tensor (in Voigt notation) for TiNi:
        [[  8.  -3.  -2.   0.  14.   0.]
         [ -3.   8.  -5.   0.  -8.   0.]
         [ -2.  -5.  10.   0.   0.   0.]
         [  0.   0.   0.  12.   0.   0.]
         [ 14.  -8.   0.   0. 116.   0.]
         [  0.   0.   0.   0.   0.  12.]]
        Symmetry: monoclinic
        """
        matrix = cls._matrixFromCrystalSymmetry(point_group=point_group, diad=diad, symmetry=symmetry, prefix=prefix,
                                                **kwargs)
        return cls(matrix, symmetry=symmetry, phase_name=phase_name)

    @classmethod
    def hexagonal(cls, *, C11=0., C12=0., C13=0., C33=0., C44=0., phase_name=None):
        """
        Create a fourth-order tensor from hexagonal symmetry.

        Parameters
        ----------
        C11, C12 , C13, C33, C44 : float
            Components of the tensor, using the Voigt notation
        phase_name : str, optional
            Phase name to display
        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        transverse_isotropic : creates a transverse-isotropic tensor from engineering parameters
        cubic : create a tensor from cubic symmetry
        tetragonal : create a tensor from tetragonal symmetry
        """
        return cls.fromCrystalSymmetry(symmetry='hexagonal', C11=C11, C12=C12, C13=C13, C33=C33, C44=C44,
                                       phase_name=phase_name, prefix='C')

    @classmethod
    def trigonal(cls, *, C11=0., C12=0., C13=0., C14=0., C33=0., C44=0., C15=0., phase_name=None):
        """
        Create a fourth-order tensor from trigonal symmetry.

        Parameters
        ----------
        C11, C12, C13, C14, C33, C44 : float
            Components of the tensor, using the Voigt notation
        C15 : float, optional
            C15 component of the tensor, only used for point groups 3 and -3.
        phase_name : str, optional
            Phase name to display
        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        tetragonal : create a tensor from tetragonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls.fromCrystalSymmetry(point_group='3', C11=C11, C12=C12, C13=C13, C14=C14, C15=C15,
                                       C33=C33, C44=C44, phase_name=phase_name, prefix='C')

    @classmethod
    def tetragonal(cls, *, C11=0., C12=0., C13=0., C33=0., C44=0., C16=0., C66=0., phase_name=None):
        """
        Create a fourth-order tensor from tetragonal symmetry.

        Parameters
        ----------
        C11,  C12, C13, C33, C44, C66 : float
            Components of the tensor, using the Voigt notation
        C16 : float, optional
            C16 component in Voigt notation (for point groups 4, -4 and 4/m only)
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        trigonal : create a tensor from trigonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls.fromCrystalSymmetry(point_group='4', C11=C11, C12=C12, C13=C13, C16=C16,
                                       C33=C33, C44=C44, C66=C66, phase_name=phase_name, prefix='C')

    @classmethod
    def cubic(cls, *, C11=0., C12=0., C44=0., phase_name=None):
        """
        Create a fourth-order tensor from cubic symmetry.

        Parameters
        ----------
        C11 , C12, C44 : float
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        hexagonal : create a tensor from hexagonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls.fromCrystalSymmetry(symmetry='cubic', C11=C11, C12=C12, C44=C44, phase_name=phase_name, prefix='C')

    @classmethod
    def orthorhombic(cls, *, C11=0., C12=0., C13=0., C22=0., C23=0., C33=0., C44=0., C55=0., C66=0., phase_name=None):
        """
        Create a fourth-order tensor from orthorhombic symmetry.

        Parameters
        ----------
        C11, C12, C13, C22, C23, C33, C44, C55, C66 : float
            Components of the tensor, using the Voigt notation
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        monoclinic : create a tensor from monoclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls.fromCrystalSymmetry(symmetry='orthorhombic',
                                       C11=C11, C12=C12, C13=C13, C22=C22, C23=C23, C33=C33, C44=C44, C55=C55, C66=C66,
                                       phase_name=phase_name, prefix='C')

    @classmethod
    def monoclinic(cls, *, C11=0., C12=0., C13=0., C22=0., C23=0., C33=0., C44=0., C55=0., C66=0.,
                   C15=None, C25=None, C35=None, C46=None,
                   C16=None, C26=None, C36=None, C45=None,
                   phase_name=None):
        """
        Create a fourth-order tensor from monoclinic symmetry. It automatically detects whether the components are given
        according to the Y or Z diad, depending on the input arguments.

        For Diad || y, C15, C25, C35 and C46 must be provided.
        For Diad || z, C16, C26, C36 and C45 must be provided.

        Parameters
        ----------
        C11, C12 , C13, C22, C23, C33, C44, C55, C66 : float
            Components of the tensor, using the Voigt notation
        C15 : float, optional
            C15 component of the tensor (if Diad || y)
        C25 : float, optional
            C25 component of the tensor (if Diad || y)
        C35 : float, optional
            C35 component of the tensor (if Diad || y)
        C46 : float, optional
            C46 component of the tensor (if Diad || y)
        C16 : float, optional
            C16 component of the tensor (if Diad || z)
        C26 : float, optional
            C26 component of the tensor (if Diad || z)
        C36 : float, optional
            C36 component of the tensor (if Diad || z)
        C45 : float, optional
            C45 component of the tensor (if Diad || z)
        phase_name : str, optional
            Name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        triclinic : create a tensor from triclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        diad_y = not (None in (C15, C25, C35, C46))
        diad_z = not (None in (C16, C26, C36, C45))
        if diad_y and diad_z:
            raise KeyError('Ambiguous diad. Provide either C15, C25, C35 and C46; or C16, C26, C36 and C45')
        elif diad_y:
            return cls.fromCrystalSymmetry(symmetry='monoclinic', diad='y',
                                           C11=C11, C12=C12, C13=C13, C22=C22, C23=C23, C33=C33, C44=C44, C55=C55,
                                           C66=C66,
                                           C15=C15, C25=C25, C35=C35, C46=C46, phase_name=phase_name, prefix='C')
        elif diad_z:
            return cls.fromCrystalSymmetry(symmetry='monoclinic', diad='z',
                                           C11=C11, C12=C12, C13=C13, C22=C22, C23=C23, C33=C33, C44=C44, C55=C55,
                                           C66=C66,
                                           C16=C16, C26=C26, C36=C36, C45=C45, phase_name=phase_name, prefix='C')
        else:
            raise KeyError('For monoclinic symmetry, one should provide either C15, C25, C35 and C46, '
                           'or C16, C26, C36 and C45.')

    @classmethod
    def triclinic(cls, C11=0., C12=0., C13=0., C14=0., C15=0., C16=0.,
                  C22=0., C23=0., C24=0., C25=0., C26=0.,
                  C33=0., C34=0., C35=0., C36=0.,
                  C44=0., C45=0., C46=0.,
                  C55=0., C56=0.,
                  C66=0., phase_name=None):
        """

        Parameters
        ----------
        C11 , C12 , C13 , C14 , C15 , C16 , C22 , C23 , C24 , C25 , C26 , C33 , C34 , C35 , C36 , C44 , C45 , C46 , C55 , C56 , C66 : float
            Components of the tensor
        phase_name : str, optional
            Name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        monoclinic : create a tensor from monoclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        matrix = np.array([[C11, C12, C13, C14, C15, C16],
                           [C12, C22, C23, C24, C25, C26],
                           [C13, C23, C33, C34, C35, C36],
                           [C14, C24, C34, C44, C45, C46],
                           [C15, C25, C35, C45, C55, C56],
                           [C16, C26, C36, C46, C56, C66]])
        return cls(matrix, phase_name=phase_name)

    def save_to_txt(self, filename, matrix_only=False):
        """
        Save the tensor to a text file.

        Parameters
        ----------
        filename : str
            Filename to save the tensor to.
        matrix_only : bool, False
            If true, only the components of tje stiffness tensor is saved (no data about phase nor symmetry)

        See Also
        --------
        from_txt_file : create a tensor from text file

        """
        with open(filename, 'w') as f:
            if not matrix_only:
                if self.phase_name is not None:
                    f.write(f"Phase Name: {self.phase_name}\n")
                f.write(f"Symmetry: {self.symmetry}\n")
            for row in self.matrix:
                f.write("  " + "  ".join(f"{value:8.2f}" for value in row) + "\n")

    @classmethod
    def from_txt_file(cls, filename):
        """
        Load the tensor from a text file.

        The two first lines can have data about phase name and symmetry, but this is not mandatory.

        Parameters
        ----------
        filename : str
            Filename to load the tensor from.

        Returns
        -------
        SymmetricTensor
            The reconstructed tensor read from the file.

        See Also
        --------
        save_to_txt : create a tensor from text file

        """
        with open(filename, 'r') as f:
            lines = f.readlines()

        # Initialize defaults
        phase_name = None
        symmetry = 'Triclinic'
        matrix_start_index = 0

        # Parse phase name if available
        if lines and lines[0].startswith("Phase Name:"):
            phase_name = lines[0].split(": ", 1)[1].strip()
            matrix_start_index += 1

        # Parse symmetry if available
        if len(lines) > matrix_start_index and lines[matrix_start_index].startswith("Symmetry:"):
            symmetry = lines[matrix_start_index].split(": ", 1)[1].strip()
            matrix_start_index += 1

        # Parse matrix
        matrix = np.loadtxt(lines[matrix_start_index:])

        # Return the reconstructed object
        return cls(matrix, phase_name=phase_name, symmetry=symmetry)

    def __getitem__(self, item):
        if self.orientations is None:
            raise IndexError('The tensor has no orientation, therefore it cannot be indexed.')
        else:
            return self._unrotate()*self.orientations[item]

class StiffnessTensor(SymmetricTensor):
    """
    Class for manipulating fourth-order stiffness tensors.
    """
    tensor_name = 'Stiffness'
    C11_C12_factor = 0.5

    def __init__(self, S, check_positive_definite=True, **kwargs):
        super().__init__(S, check_positive_definite=check_positive_definite, **kwargs)

    def __mul__(self, other):
        if isinstance(other, StrainTensor):
            new_tensor = super().__mul__(other)
            return StressTensor(new_tensor.matrix)
        elif isinstance(other, StressTensor):
            raise ValueError('You cannot multiply a stiffness tensor with a Stress tensor.')
        else:
            return super().__mul__(other)

    def inv(self):
        """
        Compute the reciprocal compliance tensor

        Returns
        -------
        ComplianceTensor
            Reciprocal tensor
        """
        C = np.linalg.inv(self.matrix)
        return ComplianceTensor(C, symmetry=self.symmetry, phase_name=self.phase_name, orientations=self.orientations)

    @property
    def Young_modulus(self):
        """
        Directional Young's modulus

        Returns
        -------
        SphericalFunction
            Young's modulus
        """

        def compute_young_modulus(n):
            eps = _compute_unit_strain_along_direction(self, n, n)
            return 1 / eps

        return SphericalFunction(compute_young_modulus)

    @property
    def shear_modulus(self):
        """
        Directional shear modulus

        Returns
        -------
        HyperSphericalFunction
            Shear modulus
        """

        def compute_shear_modulus(m, n):
            eps = _compute_unit_strain_along_direction(self, m, n)
            return 1 / (4 * eps)

        return HyperSphericalFunction(compute_shear_modulus)

    @property
    def Poisson_ratio(self):
        """
        Directional Poisson's ratio

        Returns
        -------
        HyperSphericalFunction
            Poisson's ratio
        """

        def compute_PoissonRatio(m, n):
            eps1 = _compute_unit_strain_along_direction(self, m, m)
            eps2 = _compute_unit_strain_along_direction(self, m, n, direction='transverse')
            return -eps2 / eps1

        return HyperSphericalFunction(compute_PoissonRatio)

    @property
    def linear_compressibility(self):
        """
        Compute the directional linear compressibility.

        Returns
        -------
        SphericalFunction
            Directional linear compressibility

        See Also
        --------
        bulk_modulus : bulk modulus of the material
        """
        def compute_linear_compressibility(n):
            return _compute_unit_strain_along_direction(self, n, n, direction='spherical')

        return SphericalFunction(compute_linear_compressibility)

    @property
    def bulk_modulus(self):
        """
        Compute the bulk modulus of the material

        Returns
        -------
        float
            Bulk modulus

        See Also
        --------
        linear_compressibility : directional linear compressibility
        """
        return self.inv().bulk_modulus


    def Voigt_average(self):
        """
        Compute the Voigt average of the stiffness tensor. If the tensor contains no orientation, we assume isotropic
        behaviour. Otherwise, the mean is computed over all orientations.

        Returns
        -------
        StiffnessTensor
            Voigt average of stiffness tensor

        See Also
        --------
        Reuss_average : compute the Reuss average
        Hill_average : compute the Voigt-Reuss-Hill average
        average : generic function for calling either the Voigt, Reuss or Hill average
        """
        if self.orientations is None:
            c = self.matrix
            C11 = (c[0, 0] + c[1, 1] + c[2, 2]) / 5 \
                  + (c[0, 1] + c[0, 2] + c[1, 2]) * 2 / 15 \
                  + (c[3, 3] + c[4, 4] + c[5, 5]) * 4 / 15
            C12 = (c[0, 0] + c[1, 1] + c[2, 2]) / 15 \
                  + (c[0, 1] + c[0, 2] + c[1, 2]) * 4 / 15 \
                  - (c[3, 3] + c[4, 4] + c[5, 5]) * 2 / 15
            C44 = (c[0, 0] + c[1, 1] + c[2, 2] - c[0, 1] - c[0, 2] - c[1, 2]) / 15 + (c[3, 3] + c[4, 4] + c[5, 5]) / 5
            mat = _isotropic_matrix(C11, C12, C44)
            return StiffnessTensor(mat, symmetry='isotropic', phase_name=self.phase_name)
        else:
            return self._orientation_average()

    def Reuss_average(self):
        """
        Compute the Reuss average of the stiffness tensor. If the tensor contains no orientation, we assume isotropic
        behaviour. Otherwise, the mean is computed over all orientations.

        Returns
        -------
        StiffnessTensor
            Reuss average of stiffness tensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Hill_average : compute the Voigt-Reuss-Hill average
        average : generic function for calling either the Voigt, Reuss or Hill average
        """
        return self.inv().Reuss_average().inv()

    def Hill_average(self):
        """
        Compute the (Voigt-Reuss-)Hill average of the stiffness tensor. If the tensor contains no orientation, we assume
        isotropic behaviour. Otherwise, the mean is computed over all orientations.

        Returns
        -------
        StiffnessTensor
            Voigt-Reuss-Hill average of tensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Reuss_average : compute the Reuss average
        average : generic function for calling either the Voigt, Reuss or Hill average
        """
        Reuss = self.Reuss_average()
        Voigt = self.Voigt_average()
        return (Reuss + Voigt) * 0.5

    def average(self, method):
        """
        Compute either the Voigt, Reuss, or Hill average of the stiffness tensor.

        This function is just a shortcut for Voigt_average(), Reuss_average(), or Hill_average() and Hill_average().

        Parameters
        ----------
        method : str {'Voigt', 'Reuss', 'Hill'}
        Method to use to compute the average.

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Reuss_average : compute the Reuss average
        Hill_average : compute the Voigt-Reuss-Hill average
        """
        method = method.capitalize()
        if method in ('Voigt', 'Reuss', 'Hill'):
            fun = getattr(self, method + '_average')
            return fun()
        else:
            raise NotImplementedError('Only Voigt, Reus, and Hill are implemented.')


    @classmethod
    def isotropic(cls, E=None, nu=None, lame1=None, lame2=None, phase_name=None):
        """
        Create an isotropic stiffness tensor from two elasticity coefficients, namely: E, nu, lame1, or lame2. Exactly
        two of these coefficients must be provided.

        Parameters
        ----------
        E : float, None
            Young modulus
        nu : float, None
            Poisson ratio
        lame1 : float, None
            First Lamé coefficient
        lame2 : float, None
            Second Lamé coefficient
        phase_name : str, None
            Name to print

        Returns
        -------
            Corresponding isotropic stiffness tensor

        See Also
        --------
        transverse_isotropic : create a transverse-isotropic tensor

        Examples
        --------
        On can check that the shear modulus for steel is around 82 GPa:

        >>> from Elasticipy.FourthOrderTensor import StiffnessTensor
        >>> C=StiffnessTensor.isotropic(E=210e3, nu=0.28)
        >>> C.shear_modulus
        Hyperspherical function
        Min=82031.24999999997, Max=82031.25000000006
        """
        argument_vector = np.array([E, nu, lame1, lame2])
        if np.count_nonzero(argument_vector) != 2:
            raise ValueError("Exactly two values are required among E, nu, lame1 and lame2.")
        if E is not None:
            if nu is not None:
                lame1 = E * nu / ((1 + nu) * (1 - 2 * nu))
                lame2 = E / (1 + nu) / 2
            elif lame1 is not None:
                R = np.sqrt(E ** 2 + 9 * lame1 ** 2 + 2 * E * lame1)
                lame2 = (E - 3 * lame1 + R) / 4
            elif lame2 is not None:
                lame1 = lame2 * (E - 2 * lame2) / (3 * lame2 - E)
            else:
                raise ValueError('Either nu, lame1 or lame2 must be provided.')
        elif nu is not None:
            if lame1 is not None:
                lame2 = lame1 * (1 - 2 * nu) / (2 * nu)
            elif lame2 is not None:
                lame1 = 2 * lame2 * nu / (1 - 2 * nu)
            else:
                raise ValueError('Either lame1 or lame2 must be provided.')
        C11 = lame1 + 2 * lame2
        C12 = lame1
        C44 = lame2
        matrix = _isotropic_matrix(C11, C12, C44)
        return StiffnessTensor(np.array(matrix), symmetry='isotropic', phase_name=phase_name)

    @classmethod
    def orthotropic(cls, *, Ex, Ey, Ez, nu_yx, nu_zx, nu_zy, Gxy, Gxz, Gyz, **kwargs):
        """
        Create a stiffness tensor corresponding to orthotropic symmetry, given the engineering constants.

        Parameters
        ----------
        Ex : float
            Young modulus along the x axis
        Ey : float
            Young modulus along the y axis
        Ez : float
            Young modulus along the z axis
        nu_yx : float
            Poisson ratio between x and y axes
        nu_zx : float
            Poisson ratio between x and z axes
        nu_zy : float
            Poisson ratio between y and z axes
        Gxy : float
            Shear modulus in the x-y plane
        Gxz : float
            Shear modulus in the x-z plane
        Gyz : float
            Shear modulus in the y-z plane
        kwargs : dict, optional
            Keyword arguments to pass to the StiffnessTensor constructor

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        transverse_isotropic : create a stiffness tensor for transverse-isotropic symmetry
        """
        tri_sup = np.array([[1 / Ex, -nu_yx / Ey, -nu_zx / Ez, 0, 0, 0],
                            [0, 1 / Ey, -nu_zy / Ez, 0, 0, 0],
                            [0,      0,           1 / Ez,       0,          0,          0],
                            [0,      0,           0,            1 / Gyz,    0,          0],
                            [0,      0,           0,            0,          1 / Gxz,    0],
                            [0,      0,           0,            0,          0,          1 / Gxy]])
        S = tri_sup + np.tril(tri_sup.T, -1)
        return StiffnessTensor(np.linalg.inv(S), symmetry='orthotropic', **kwargs)

    @classmethod
    def transverse_isotropic(cls, *, Ex, Ez, nu_yx, nu_zx, Gxz, **kwargs):
        """
        Create a stiffness tensor corresponding to the transverse isotropic symmetry, given the engineering constants.

        Parameters
        ----------
        Ex : float
            Young modulus along the x axis
        Ez : float
            Young modulus along the y axis
        nu_yx : float
            Poisson ratio between x and y axes
        nu_zx : float
            Poisson ratio between x and z axes
        Gxz : float
            Shear modulus in the x-z plane
        kwargs : dict
            Keyword arguments to pass to the StiffnessTensor constructor

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        orthotropic : create a stiffness tensor for orthotropic symmetry
        """
        Gxy = Ex / (2 * (1 + nu_yx))
        C = StiffnessTensor.orthotropic(Ex=Ex, Ey=Ex, Ez=Ez,
                                        nu_yx=nu_yx, nu_zx=nu_zx, nu_zy=nu_zx,
                                        Gxy=Gxy, Gxz=Gxz, Gyz=Gxz, **kwargs)
        C.symmetry = 'transverse-isotropic'
        return C

    def Christoffel_tensor(self, u):
        """
        Create the Christoffel tensor along a given direction, or set or directions.

        Parameters
        ----------
        u : list or np.ndarray
            3D direction(s) to compute the Christoffel tensor along with

        Returns
        -------
        Gamma : np.ndarray
            Array of Christoffel tensor(s). if u is a list of directions, Gamma[i] is the Christoffel tensor for
            direction  u[i].

        See Also
        --------
        wave_velocity : computes the p- and s-wave velocities.

        Notes
        -----
        For a given stiffness tensor **C** and a given unit vector **u**, the Christoffel tensor is defined as [2]_ :

            .. math:: M_{ij} = C_{iklj}.u_k.u_l

        """
        u_vec = np.atleast_2d(u)
        u_vec = (u_vec.T / np.linalg.norm(u_vec, axis=1)).T
        return np.einsum('inmj,pn,pm->pij', self.full_tensor(), u_vec, u_vec)

    def wave_velocity(self, rho):
        """
        Compute the wave velocities, given the mass density.

        Parameters
        ----------
        rho : float
            mass density. Its unit must be consistent with that of the stiffness tensor. See notes for hints.

        See Also
        --------
        ChristoffelTensor : Computes the Christoffel tensor along a given direction

        Returns
        -------
        c_p : SphericalFunction
            Velocity of the primary (compressive) wave
        c_s1 : SphericalFunction
            Velocity of the fast secondary (shear) wave
        c_s2 : SphericalFunction
            Velocity of the slow secondary (shear) wave

        Notes
        -----
        The estimation of the wave velocities is made by finding the eigenvalues of the Christoffel tensor [2]_.

        One should double-check the units. The table below provides hints about the unit you get, depending on the units
        you use for stiffness and the mass density:

        +-----------------+--------------+------------+-----------------------+
        | Stiffness       | Mass density | Velocities | Notes                 |
        +=================+==============+============+=======================+
        | Pa (N/m²)       | kg/m³        | m/s        | SI units              |
        +-----------------+--------------+------------+-----------------------+
        | GPa (10⁹ Pa)    | kg/dm³       | km/s       | Conversion factor     |
        +-----------------+--------------+------------+-----------------------+
        | GPa (10³ N/mm²) | kg/mm³       | m/s        | Consistent units      |
        +-----------------+--------------+------------+-----------------------+
        | MPa (10⁶ Pa)    | kg/m³        | km/s       | Conversion factor     |
        +-----------------+--------------+------------+-----------------------+
        | MPa (10³ N/mm²) | g/mm³        | m/s        | Consistent units      |
        +-----------------+--------------+------------+-----------------------+

        References
        ----------
        .. [2] J. W. Jaeken, S. Cottenier, Solving the Christoffel equation: Phase and group velocities, Computer Physics
               Communications (207), 2016, https://doi.org/10.1016/j.cpc.2016.06.014.

        """

        def make_fun(index):
            def fun(n):
                Gamma = self.Christoffel_tensor(n)
                eig, _ = np.linalg.eig(Gamma)
                if index == 0:
                    eig_of_interest = np.max(eig, axis=-1)
                elif index == 1:
                    eig_of_interest = np.median(eig, axis=-1)
                else:
                    eig_of_interest = np.min(eig, axis=-1)
                return np.sqrt(eig_of_interest / rho)

            return fun

        return [SphericalFunction(make_fun(i)) for i in range(3)]

    @classmethod
    def from_MP(cls, ids, api_key=None):
        """
        Import stiffness tensor(s) from the Materials Project API, given their material ids.

        You need to register to `<https://materialsproject.org>`_ first to get an API key. This key can be explicitly
        passed as an argument (see below), or provided as an environment variable named MP_API_KEY.

        Parameters
        ----------
        ids : str or list of str
            ID(s) of the material to import (e.g. "mp-1048")
        api_key : str, optional
            API key to the Materials Project API. If not provided, it should be available as the API_KEY environment
            variable.

        Returns
        -------
        list of StiffnessTensor
            If one of the requested material ids was not found, the corresponding value in the list will be None.
        """
        try:
            from mp_api.client import MPRester
        except ImportError:
            raise ModuleNotFoundError('mp_api module is required for this function.')
        if type(ids) is str:
            Cdict = dict.fromkeys([ids])
        else:
            Cdict = dict.fromkeys(ids)
        with MPRester(api_key=api_key) as mpr:
            elasticity_doc = mpr.materials.elasticity.search(material_ids=ids)
            for material in elasticity_doc:
                key = str(material.material_id)
                if material.elastic_tensor is not None:
                    matrix = material.elastic_tensor.ieee_format
                    symmetry = material.symmetry.crystal_system.value
                    phase_name = material.formula_pretty
                    C = StiffnessTensor(matrix, symmetry=symmetry, phase_name=phase_name)
                else:
                    C = None
                Cdict[key] = C
            if elasticity_doc:
                if isinstance(ids, str):
                    return C
                else:
                    return [Cdict[id] for id in ids]
            else:
                return None

    @classmethod
    def weighted_average(cls, Cs, volume_fractions, method):
        """
        Compute the weighted average of a list of stiffness tensors, with respect to a given method (Voigt, Reuss or
        Hill).

        Parameters
        ----------
        Cs : list of StiffnessTensor or list of ComplianceTensor or tuple of StiffnessTensor or tuple of ComplianceTensor
            Series of tensors to compute the average from
        volume_fractions : iterable of floats
            Volume fractions of each phase
        method : str, {'Voigt', 'Reuss', 'Hill'}
            Method to use. It can be 'Voigt', 'Reuss', or 'Hill'.

        Returns
        -------
        StiffnessTensor
            Average tensor
        """
        if np.all([isinstance(a, ComplianceTensor) for a in Cs]):
            Cs = [C.inv() for C in Cs]
        if np.all([isinstance(a, StiffnessTensor) for a in Cs]):
            C_stack = np.array([C.matrix for C in Cs])
            method = method.capitalize()
            if method == 'Voigt':
                C_avg = np.average(C_stack, weights=volume_fractions, axis=0)
                return StiffnessTensor(C_avg)
            elif method == 'Reuss':
                S_stack = np.linalg.inv(C_stack)
                S_avg = np.average(S_stack, weights=volume_fractions, axis=0)
                return StiffnessTensor(np.linalg.inv(S_avg))
            elif method == 'Hill':
                C_voigt = cls.weighted_average(Cs, volume_fractions, 'Voigt')
                C_reuss = cls.weighted_average(Cs, volume_fractions, 'Reuss')
                return (C_voigt + C_reuss) * 0.5
            else:
                raise ValueError('Method must be either Voigt, Reuss or Hill.')
        else:
            raise ValueError('The first argument must be either a list of ComplianceTensors or '
                             'a list of StiffnessTensor.')

    @property
    def universal_anisotropy(self):
        """
        Compute the universal anisotropy factor.

        The larger the value, the more likely the material will behave in an anisotropic way.

        Returns
        -------
        float
            The universal anisotropy factor.

        Notes
        -----
        The universal anisotropy factor is defined as [3]_:

        .. math::

            5\\frac{G_v}{G_r} + \\frac{K_v}{K_r} - 6

        References
        ----------
        .. [3] S. I. Ranganathan and M. Ostoja-Starzewski, Universal Elastic Anisotropy Index,
           *Phys. Rev. Lett.*, 101(5), 055504, 2008. https://doi.org/10.1103/PhysRevLett.101.055504
        """
        C = self._unrotate()    # Ensure that the averages do not use the orientations
        Cvoigt = C.Voigt_average()
        Creuss = C.Reuss_average()
        Gv = Cvoigt.matrix[3, 3]
        Gr = Creuss.matrix[3, 3]
        Kv = Cvoigt.bulk_modulus
        Kr = Creuss.bulk_modulus
        return 5 * Gv / Gr + Kv / Kr - 6

    @property
    def Zener_ratio(self):
        """
        Compute the Zener ratio (Z). Only valid for cubic symmetry.

        It is only valid for cubic and isotropic symmetry. Will return NaN for other symmetries.

        Returns
        -------
        float
            Zener ratio (NaN is the symmetry is not cubic)

        Notes
        -----
        The Zener ratio is defined as:

        .. math::

                Z=\\frac{ 2C_{44} }{C11 - C12}

        See Also
        --------
        universal_anisotropy : compute the universal anisotropy factor
        """
        if self.symmetry == 'isotropic':
            return 1.0
        elif self.symmetry == 'cubic':
            return 2 * self.C44 / (self.C11 - self.C12)
        else:
            return np.nan

    def to_pymatgen(self):
        """
        Convert the stiffness tensor (from Elasticipy) to Python Materials Genomics (Pymatgen) format.

        Returns
        -------
        pymatgen.analysis.elasticity.elastic.ElasticTensor
            Stiffness tensor for pymatgen
        """
        try:
            from pymatgen.analysis.elasticity import elastic as matgenElast
        except ImportError:
            raise ModuleNotFoundError('pymatgen module is required for this function.')
        return matgenElast.ElasticTensor(self.full_tensor())

class ComplianceTensor(StiffnessTensor):
    """
    Class for manipulating compliance tensors
    """
    tensor_name = 'Compliance'
    voigt_map = np.array([[1., 1., 1., 2., 2., 2.],
                          [1., 1., 1., 2., 2., 2.],
                          [1., 1., 1., 2., 2., 2.],
                          [2., 2., 2., 4., 4., 4.],
                          [2., 2., 2., 4., 4., 4.],
                          [2., 2., 2., 4., 4., 4.]])
    C11_C12_factor = 2.0
    component_prefix = 'S'
    C46_C56_factor = 2.0

    def __init__(self, C, check_positive_definite=True, **kwargs):
        super().__init__(C, check_positive_definite=check_positive_definite, **kwargs)

    def __mul__(self, other):
        if isinstance(other, StressTensor):
            return StrainTensor(self * other.matrix)
        elif isinstance(other, StrainTensor):
            raise ValueError('You cannot multiply a compliance tensor with Strain tensor.')
        else:
            return super().__mul__(other)

    def inv(self):
        """
        Compute the reciprocal stiffness tensor

        Returns
        -------
        StiffnessTensor
            Reciprocal tensor
        """
        S = np.linalg.inv(self.matrix)
        return StiffnessTensor(S, symmetry=self.symmetry, phase_name=self.phase_name, orientations=self.orientations)

    def Reuss_average(self):
        if self.orientations is None:
            s = self.matrix
            S11 = (s[0, 0] + s[1, 1] + s[2, 2]) / 5 \
                  + (s[0, 1] + s[0, 2] + s[1, 2]) * 2 / 15 \
                  + (s[3, 3] + s[4, 4] + s[5, 5]) * 1 / 15
            S12 = (s[0, 0] + s[1, 1] + s[2, 2]) / 15 \
                  + (s[0, 1] + s[0, 2] + s[1, 2]) * 4 / 15 \
                  - (s[3, 3] + s[4, 4] + s[5, 5]) * 1 / 30
            S44 = ((s[0, 0] + s[1, 1] + s[2, 2] - s[0, 1] - s[0, 2] - s[1, 2]) * 4 / 15 +
                   (s[3, 3] + s[4, 4] + s[5, 5]) / 5)
            mat = _isotropic_matrix(S11, S12, S44)
            return ComplianceTensor(mat, symmetry='isotropic', phase_name=self.phase_name)
        else:
            return self._orientation_average()

    def Voigt_average(self):
        return self.inv().Voigt_average().inv()

    def Hill_average(self):
        return self.inv().Hill_average().inv()

    @classmethod
    def isotropic(cls, E=None, nu=None, lame1=None, lame2=None, phase_name=None):
        return super().isotropic(E=E, nu=nu, lame1=lame1, lame2=lame2, phase_name=None).inv()

    @classmethod
    def orthotropic(cls, *args, **kwargs):
        return super().orthotropic(*args, **kwargs).inv()

    @classmethod
    def transverse_isotropic(cls, *args, **kwargs):
        return super().transverse_isotropic(*args, **kwargs).inv()
    
    @classmethod
    def weighted_average(cls, *args):
        return super().weighted_average(*args).inv()

    @property
    def bulk_modulus(self):
        return 1/np.sum(self.matrix[0:3,0:3])

    @property
    def universal_anisotropy(self):
        """
        Compute the universal anisotropy factor.

        It is actually an alias for inv().universal_anisotropy.

        Returns
        -------
        float
            Universal anisotropy factor
        """
        return self.inv().universal_anisotropy

    def to_pymatgen(self):
        """
        Convert the compliance tensor (from Elasticipy) to Python Materials Genomics (Pymatgen) format.

        Returns
        -------
        pymatgen.analysis.elasticity.elastic.Compliance
            Compliance tensor for pymatgen
        """
        try:
            from pymatgen.analysis.elasticity import elastic as matgenElast
        except ImportError:
            raise ModuleNotFoundError('pymatgen module is required for this function.')
        return matgenElast.ComplianceTensor(self.full_tensor())



import numpy as np
from Elasticipy.SecondOrderTensor import SymmetricSecondOrderTensor


class StrainTensor(SymmetricSecondOrderTensor):
    """
    Class for manipulating symmetric strain tensors or arrays of symmetric strain tensors.

    """
    name = 'Strain tensor'
    voigt_map = [1, 1, 1, 2, 2, 2]

    def principal_strains(self):
        """
        Values of the principals strains.

        If the tensor array is of shape [m,n,...], the results will be of shape [m,n,...,3].

        Returns
        -------
        np.ndarray
            Principal strain values
        """
        return self.eigvals()

    def volumetric_strain(self):
        """
        Volumetric change (1st invariant of the strain tensor)

        Returns
        -------
        np.array or float
            Volumetric change
        """
        return self.I1

    def eq_strain(self):
        return np.sqrt(3/2 * self.J2)

    def elastic_energy(self, stress):
        """
        Compute the elastic energy.

        Parameters
        ----------
        stress : StressTensor
            Corresponding stress tensor

        Returns
        -------
        Volumetric elastic energy
        """
        return 0.5 * self.ddot(stress)


class StressTensor(SymmetricSecondOrderTensor):
    """
    Class for manipulating stress tensors or arrays of stress tensors.
    """
    name = 'Stress tensor'

    def principal_stresses(self):
        """
        Values of the principals stresses.

        If the tensor array is of shape [m,n,...], the results will be of shape [m,n,...,3].

        Returns
        -------
        np.ndarray
            Principal stresses
        """
        return self.eigvals()

    def vonMises(self):
        """
        von Mises equivalent stress.

        Returns
        -------
        np.ndarray or float
            von Mises equivalent stress

        See Also
        --------
        Tresca : Tresca equivalent stress
        """
        return np.sqrt(3 * self.J2)

    def Tresca(self):
        """
        Tresca(-Guest) equivalent stress.

        Returns
        -------
        np.ndarray or float
            Tresca equivalent stress

        See Also
        --------
        vonMises : von Mises equivalent stress
        """
        ps = self.principal_stresses()
        return ps[...,-1] - ps[...,0]

    def hydrostaticPressure(self):
        """
        Hydrostatic pressure

        Returns
        -------
        np.ndarray or float

        See Also
        --------
        sphericalPart : spherical part of the stress
        """
        return -self.I1/3

    def elastic_energy(self, strain):
        """
        Compute the elastic energy.

        Parameters
        ----------
        strain : StrainTensor
            Corresponding strain tensor

        Returns
        -------
        Volumetric elastic energy
        """
        return 0.5 * self.ddot(strain)


from Elasticipy.FourthOrderTensor import StiffnessTensor
import matplotlib as mpl
mpl.use('Qt5Agg')   # Ensure interactive plot
from matplotlib import pyplot as plt

C = StiffnessTensor.fromCrystalSymmetry(symmetry='orthorombic', phase_name='forsterite',
                                        C11=320, C12=68.2, C13=71.6,
                                        C22=196.5, C23=76.8, C33=233.5, C44=64, C55=77, C66=78.7)
rho = 3.355

cp, cs_fast, cs_slow = C.wave_velocity(rho)
print(cp)

fig, _ =cp.plot_as_pole_figure(subplot_args=(131,), title='p wave', show=False)
cs_fast.plot_as_pole_figure(subplot_args=(132,), title='s wave (fast)', fig=fig, show=False)
cs_slow.plot_as_pole_figure(subplot_args=(133,), title='s wave (slow)', fig=fig, show=True)



# Configuration file for the Sphinx documentation builder.
#
import os
import sys
src_path = os.path.abspath('../../src/')
sys.path.insert(0, src_path)
print(f"Chemin ajouté au PYTHONPATH : {sys.path}")

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Elasticipy'
copyright = '2024, Dorian Depriester'
author = 'Dorian Depriester'
release = '1.0.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
   'sphinx_rtd_theme',
   'sphinx.ext.doctest',
   'sphinx.ext.autodoc',
   'sphinx.ext.autosummary',
   'sphinx.ext.napoleon',
   'sphinx.ext.githubpages',
   'sphinx.ext.autosectionlabel',
   'sphinx.ext.mathjax',
   'sphinx.ext.linkcode']
templates_path = ['_templates']
exclude_patterns = []

language = 'english'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['../_static']
numpydoc_class_members_toctree = False
autoclass_content = 'both'

def linkcode_resolve(domain, info):
    if domain != 'py':
        return None
    if not info['module']:
        return None
    filename = info['module'].replace('.', '/')
    return f"https://github.com/DorianDepriester/Elasticipy/blob/main/src/{filename}.py"



import unittest
from matplotlib import pyplot as plt
import numpy as np

from Elasticipy.FourthOrderTensor import StiffnessTensor
from pytest import approx

from Elasticipy.SphericalFunction import HyperSphericalFunction

C = StiffnessTensor.cubic(C11=186, C12=134, C44=77)
E = C.Young_modulus # SphericalFunction
E_mean = 126.28067650635076
E_std = 31.58751357560234
G = C.shear_modulus # HypersphericalFunction
G_mean = 47.07147379585229
G_std = 14.14600864639266
SEED = 123  # Used for Monte Carlo integrations (e.g. for G.mean())

import re


def template_test_repr(result, class_name, expected_min, expected_max):
    # First, check if the string if correctly formated
    pattern = r'{} function\nMin=(\d+\.\d+), Max=(\d+\.\d+)'.format(class_name)
    match = re.match(pattern, result)
    assert match is not None

    # Now check the returned values
    min_val = float(match.group(1))
    max_val = float(match.group(2))
    assert expected_min == approx(min_val)
    assert expected_max == approx(max_val)


class TestSphericalFunction(unittest.TestCase):
    def test_plot3D(self):
        fig = plt.figure()
        _, ax = E.plot3D(fig=fig)
        np.testing.assert_allclose(ax.xaxis.v_interval, [-174.66981, 175.43156])
        np.testing.assert_allclose(ax.yaxis.v_interval, [-175.05069, 175.05069])
        np.testing.assert_allclose(ax.zaxis.v_interval, [-131.28802, 131.28802])

    def test_plot_xyz_section(self):
        fig = plt.figure()
        _, axs = E.plot_xyz_sections(fig=fig)
        assert axs[0].title._text == 'X-Y plane'
        assert axs[1].title._text == 'X-Z plane'
        assert axs[2].title._text == 'Y-Z plane'

    def test_plot_as_pole_figure(self):
        _, ax = E.plot_as_pole_figure(show=False)
        np.testing.assert_allclose(ax.dataLim.intervalx, [-0.01578689775673263, 6.298972204936319])
        np.testing.assert_allclose(ax.dataLim.intervaly, [-0.0160285339468867, 1.5868248607417832])

    def test_add_sub_mult_div(self):
        E_plus = E + E
        E_min = E - E
        E_mult = 2 * E
        E_plus_one = E + 1
        E_minus_one = E - 1
        E_div = E / E
        E_div_two = E / 2
        E_square = E * E
        assert E_plus.mean() == approx(2 * E_mean, rel=1e-3)
        assert E_min.mean() == approx(0)
        assert E_mult.mean() == approx(2 * E_mean, rel=1e-3)
        assert E_plus_one.mean() == approx(E_mean + 1, rel=1e-3)
        assert E_minus_one.mean() == approx(E_mean - 1, rel=1e-3)
        assert E_div.mean() == approx(1, rel=1e-3)
        assert E_div_two.mean() == approx(E_mean/2, rel=1e-3)
        assert E_square.eval([1,0,0]) == approx(E.eval([1,0,0])**2, rel=1e-3)

        expected_error = 'A Spherical function can only be added to another Spherical function or a scalar value.'
        with self.assertRaises(NotImplementedError) as context:
            _ = E + G
        self.assertEqual(str(context.exception), expected_error)

        expected_error = 'A Spherical function can only be multiplied by another Spherical function or a scalar value.'
        with self.assertRaises(NotImplementedError) as context:
            _ = E * G
        self.assertEqual(str(context.exception), expected_error)

        expected_error ='A SphericalFunction can only be divided by a scalar value of another SphericalFunction.'
        with self.assertRaises(NotImplementedError) as context:
            _ = E / G
        self.assertEqual(str(context.exception), expected_error)

    def test_mean_std(self):
        for method in ('exact', 'trapezoid', 'Monte Carlo'):
            assert E_mean == approx(E.mean(method=method, seed=SEED), rel=1e-2)
            assert E_std == approx(E.std(method=method, seed=SEED), rel=1e-2)


    def test_repr(self):
        template_test_repr(E.__repr__(), 'Spherical', 73.775, 197.50282485875343)

    def test_eval_spherical(self):
        assert E.eval_spherical([0, 0], degrees=True) == approx(E.eval([0, 0, 1]))
        assert E.eval_spherical([0, 90], degrees=True) == approx(E.eval([1, 0, 0]))
        assert E.eval_spherical([90, 90], degrees=True) == approx(E.eval([0, 1, 0]))


class TestHyperSphericalFunction(unittest.TestCase):
    def test_plot3D(self):
        fig = plt.figure()
        _, ax = G.plot3D(fig=fig)
        np.testing.assert_allclose(ax.xaxis.v_interval, [-86.363067,  87.737754])
        np.testing.assert_allclose(ax.yaxis.v_interval, [-87.435894,  87.435894])
        np.testing.assert_allclose(ax.zaxis.v_interval, [-80.208333,  80.208333])

    def test_plot_xyz_section(self):
        fig = plt.figure()
        _, axs = G.plot_xyz_sections(fig=fig)
        assert axs[0].title._text == 'X-Y plane'
        assert axs[1].title._text == 'X-Z plane'
        assert axs[2].title._text == 'Y-Z plane'

    def test_plot_as_pole_figure(self):
        _, ax = G.plot_as_pole_figure(show=False)
        np.testing.assert_allclose(ax.dataLim.intervalx, [-0.01578689775673263, 6.298972204936319])
        np.testing.assert_allclose(ax.dataLim.intervaly, [-0.0160285339468867, 1.5868248607417832])

    def test_add_sub_mult_div(self):
        G_plus = G + G
        G_min = G - G
        G_mult = 2 * G
        Gplus_one = G + 1
        G_div = G / G
        G_div_two = G / 2
        assert G_plus.mean() == approx(2 * G_mean, rel=5e-3)
        assert G_min.mean() == approx(0)
        assert G_mult.mean() == approx(2 * G_mean, rel=5e-3)
        assert Gplus_one.mean() == approx(G_mean + 1, rel=1e-3)
        assert G_div.mean() == approx(1, rel=1e-3)
        assert G_div_two.mean() == approx(G_mean/2, rel=1e-3)

    def test_mean_std(self):
        for method in ('trapezoid', 'Monte Carlo'):
            assert G_mean == approx(G.mean(method=method, seed=SEED), rel=5e-3)
            assert G_std  == approx(G.std( method=method, seed=SEED), rel=5e-2)


    def test_repr(self):
        template_test_repr(G.__repr__(), 'Hyperspherical', 26., 77)

    def test_eval_along_null_direction(self):
        directions = ([0,0,0],
                      np.array([[1,0,0], [0,1,0], [0,0,0]]))
        for direction in directions:
            with self.assertRaises(ValueError) as context:
                _ = E.eval(direction)
            self.assertEqual(str(context.exception), 'The input vector cannot be zeros')

    def test_eval_along_parallel_directions(self):
        directions = [([1, 0, 0], [1, 1, 0]), (np.array([[1, 0, 0], [0, 1, 0]]),
                                               np.array([[0, 1, 0], [0, 1, 1]]))]
        for direction in directions:
            with self.assertRaises(ValueError) as context:
                _ = G.eval(*direction)
            self.assertEqual(str(context.exception), 'The two directions must be orthogonal.')

    def test_eval_spherical(self):
        assert G.eval_spherical([0,0,0]) == approx(G.eval([0,0,1], [1,0,0]))
        assert G.eval_spherical([0, 0, 90], degrees=True) == approx(G.eval([0, 0, 1], [0, 1, 0]))

    def test_exact_mean(self):
        def fun(u,v):
            return [1.0]

        a = HyperSphericalFunction(fun)
        assert a.mean(method='exact') == approx(1.0)

if __name__ == '__main__':
    unittest.main()


import unittest
from pytest import approx
import numpy as np
from scipy.spatial.transform import Rotation

from Elasticipy.FourthOrderTensor import StiffnessTensor
import Elasticipy.StressStrainTensors as Tensors
from Elasticipy.SecondOrderTensor import SecondOrderTensor, SymmetricSecondOrderTensor, SkewSymmetricSecondOrderTensor
from Elasticipy.StressStrainTensors import StrainTensor, StressTensor
from pymatgen.analysis.elasticity import Strain as mgStrain, Stress as mgStress

Cmat = [[231, 127, 104, 0, -18, 0],
        [127, 240, 131, 0, 1, 0],
        [104, 131, 175, 0, -3, 0],
        [0, 0, 0, 81, 0, 3],
        [-18, 1, -3, 0, 11, 0],
        [0, 0, 0, 3, 0, 85]]
C = StiffnessTensor(Cmat)


class TestStressStrainTensors(unittest.TestCase):
    def test_mult_by_stiffness(self):
        """
        Test Stiffness/Strain tensors product C*eps (which stands for C:E)
        """
        tensile_dir = [1, 0, 0]
        stress = Tensors.StressTensor([[1, 0, 0],
                                       [0, 0, 0],
                                       [0, 0, 0]])
        strain = C.inv()*stress
        eps_xx = strain.C[0, 0]
        eps_yy = strain.C[1, 1]
        eps_zz = strain.C[2, 2]
        E = C.Young_modulus.eval(tensile_dir)
        nu_y = C.Poisson_ratio.eval(tensile_dir, [0, 1, 0])
        nu_z = C.Poisson_ratio.eval(tensile_dir, [0, 0, 1])
        assert eps_xx == approx(1/E)
        assert eps_yy == approx(-nu_y / E)
        assert eps_zz == approx(-nu_z / E)

    def test_rotate_tensor(self, n_oris=100):
        """
        Test the rotation of a tensor

        Parameters
        ----------
        n_oris : int
            Number of random orientations to use
        """
        random_tensor = SecondOrderTensor(np.random.random((3, 3)))
        random_oris = Rotation.random(n_oris)
        eps_rotated = random_tensor * random_oris
        for i in range(n_oris):
            rot_mat = random_oris[i].as_matrix()
            eps_matrix_th = np.matmul(np.matmul(rot_mat.T, random_tensor.matrix), rot_mat)
            np.testing.assert_almost_equal(eps_matrix_th, eps_rotated[i].matrix)

    def test_transpose_array(self):
        """
        Test transposing a tensor array
        """
        shape = (1, 2, 3)
        random_matrix = np.random.random(shape + (3, 3))
        random_tensor = SecondOrderTensor(random_matrix)
        transposed_tensor = random_tensor.T
        for i in range(shape[0]):
            for j in range(shape[1]):
                for k in range(shape[2]):
                    np.testing.assert_array_equal(random_matrix[i, j, k], transposed_tensor[k, j, i].matrix)

        # Check that transposing a single tensor has no effect
        a = SecondOrderTensor.rand()
        assert a == a.transposeArray()

    def test_transpose_tensor(self):
        """
        Test transposing a tensor array
        """
        shape = (2, 3, 4)
        tensor = SecondOrderTensor(np.random.random(shape + (3, 3)))
        tensor_transposed = tensor.transposeTensor()
        for i in range(shape[0]):
            for j in range(shape[1]):
                for k in range(shape[2]):
                    np.testing.assert_array_equal(tensor_transposed[i, j, k].matrix, tensor[i, j, k].matrix.T)


    def test_mul(self):
        """
        Test the element-wise product of tensors.
        """
        # First, multiply a SymmetricSecondOrderTensor with another one, and expect a matrix product between each
        # sliced matrix
        shape = (4, 5)
        shape = shape + (3, 3)
        matrix1 = np.random.random(shape)
        matrix2 = np.random.random(shape)
        tensor_prod = SecondOrderTensor(matrix1) * SecondOrderTensor(matrix2)
        for i in range(shape[0]):
            for j in range(shape[1]):
                mat_prod = np.matmul(matrix1[i, j], matrix2[i, j])
                np.testing.assert_array_equal(tensor_prod[i, j].matrix, mat_prod)

        # Now, multiply a SymmetricSecondOrderTensor with an array of the same shape, and expect an element-wise
        # multiplication between the sliced matrix of the tensor and the values of the array
        t = SecondOrderTensor(matrix1)
        random_array = np.random.random(t.shape)
        tensor_prod = t * random_array
        for i in range(shape[0]):
            for j in range(shape[1]):
                matrix = tensor_prod[i, j].matrix
                np.testing.assert_array_equal(matrix, t.matrix[i,j,:] * random_array[i,j])

    def test_matmul(self, length1=3, length2=4):
        """
        Test the matrix-like product of tensor arrays

        Parameters
        ----------
        length1 : int
            Length of the first array
        length2 : int
            Length of the second array
        """
        matrix1 = np.random.random((length1, 3, 3))
        matrix2 = np.random.random((length2, 3, 3))
        rand_tensor1 = SecondOrderTensor(matrix1)
        rand_tensor2 = SecondOrderTensor(matrix2)
        cross_prod_tensor = rand_tensor1.matmul(rand_tensor2)
        for i in range(0, length1):
            for j in range(0, length2):
                mat_prod = np.matmul(matrix1[i], matrix2[j])
                np.testing.assert_array_equal(cross_prod_tensor[i, j].matrix, mat_prod)

    def test_matmul_rotation(self):
        m, n = 5, 100
        random_tensor = SecondOrderTensor(np.random.random((m,) + (3, 3)))
        random_oris = Rotation.random(n)
        array = random_tensor.matmul(random_oris)
        assert array.shape == (m, n)
        for i in range(m):
            for j in range(n):
                rot_mat = random_oris[j].as_matrix()
                matrix = np.matmul(np.matmul(rot_mat.T, random_tensor[i].matrix), rot_mat)
                np.testing.assert_almost_equal(matrix, array[i,j].matrix)

    def test_statistics(self):
        """
        Test the std, min and max functions for tensor arrays.
        """
        shape = (5, 4, 3, 2)
        matrix = np.random.random(shape + (3, 3))
        tensor = SecondOrderTensor(matrix)
        mini = tensor.min()
        maxi = tensor.max()
        std = tensor.std()
        # First, check T.std()
        for i in range(0, 3):
            for j in range(0, 3):
                Cij = matrix[..., i, j].flatten()
                assert np.std(Cij) == approx(std.C[i, j])
                assert np.min(Cij) == approx(mini.C[i, j])
                assert np.max(Cij) == approx(maxi.C[i, j])
        # Then, check T.std(axis=...)
        for i in range(0, len(shape)):
            np.testing.assert_array_equal(tensor.std(axis=i).matrix, np.std(matrix, axis=i))
            np.testing.assert_array_equal(tensor.min(axis=i).matrix, np.min(matrix, axis=i))
            np.testing.assert_array_equal(tensor.max(axis=i).matrix, np.max(matrix, axis=i))

        # Now, check for single value tensors
        tensor = tensor[0,0,0,0]
        np.testing.assert_array_equal(tensor.min().matrix, tensor.matrix)
        np.testing.assert_array_equal(tensor.max().matrix, tensor.matrix)
        np.testing.assert_array_equal(tensor.std().matrix, np.zeros((3,3)))

    def test_ddot(self):
        """
        Test the ddot method.
        """
        shape = (4, 3, 2)
        matrix1 = np.random.random(shape + (3, 3))
        matrix2 = np.random.random(shape + (3, 3))
        tens1 = SecondOrderTensor(matrix1)
        tens2 = SecondOrderTensor(matrix2)
        ddot = tens1.ddot(tens2)
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    ddot_th = np.trace(np.matmul(matrix1[i,j,k].T, matrix2[i,j,k]))
                    self.assertEqual(ddot_th, ddot[i, j, k])

    def test_vonMises_Tresca(self):
        """
        Check that the Tresca and von Mises methods work well for simple tension, simple shear and hydrostatic
        pressure.
        """
        matrix = np.zeros((3, 3, 3))
        matrix[0, 0, 0] = 1  # Simple tension
        matrix[1, 1, 0] = matrix[1, 0, 1] = 1  # Simple shear
        matrix[2, np.arange(3), np.arange(3)] = -1  # Hydrostatic pressure
        stress = Tensors.StressTensor(matrix)

        vM_stress = stress.vonMises()
        vm_th = np.array([1, np.sqrt(3), 0.0])
        np.testing.assert_array_equal(vM_stress, vm_th)

        Tresca_stress = stress.Tresca()
        Tresca_th = np.array([1, 2, 0.0])
        np.testing.assert_array_equal(Tresca_stress, Tresca_th)

    def test_rotation_stiffness(self, ):
        """
        Check that the two ways to compute stress from a rotated stiffness tensor are consistent.
        """
        n_strain = 50
        n_ori = 100
        matrix = np.random.random((n_strain, 3, 3))
        eps = Tensors.StrainTensor(matrix, force_symmetry=True)
        ori = Rotation.random(n_ori)
        C_rotated = C * ori
        sigma = C_rotated * eps

        # Rotate stress and stress by their own
        eps_rot = eps.matmul(ori)
        sigma_rot2 = C * eps_rot
        sigma2 = sigma_rot2 * ori.inv()
        np.testing.assert_almost_equal(sigma.matrix, sigma2.transposeArray().matrix)

    def test_multidimensional_tensors(self, ):
        """
        Check that the shape of (C * rotations) * eps is (m, p, r, ...) if rotation.shape=(p, q,...) and
        len(rotations)=m.
        """
        shape_strain = (5, 4, 3)
        n_ori = 100
        strain = Tensors.StrainTensor.ones(shape_strain)
        ori = Rotation.random(n_ori)
        C_rotated = C * ori
        stress = C_rotated * strain
        self.assertEqual(stress.shape, (n_ori,) + shape_strain)

    def test_Voigt_notation_strain(self):
        """
        Check that the strain tensor can be reconstructed from Voigt vectors.
        Returns
        """
        a = np.random.random((3, 4, 6))
        strain = Tensors.StrainTensor.from_Voigt(a)
        for i in range(0, 3):
            for j in range(0, 4):
                for k in range(0, 6):
                    if k<3:
                        assert a[i,j,k] == strain[i,j].C[k,k]
                    elif k==3:
                        assert a[i,j,k] == 2*strain[i,j].C[1,2]
                    elif k==4:
                        assert a[i,j,k] == 2*strain[i,j].C[0,2]
                    else:
                        assert a[i,j,k] == 2*strain[i,j].C[0,1]

    def test_Voigt_notation_stress(self):
        """
        Check that the stress tensor can be reconstructed from Voigt vectors.
        """
        a = np.random.random((3, 4 , 6))
        stress = Tensors.StressTensor.from_Voigt(a)
        for i in range(0, 3):
            for j in range(0, 4):
                for k in range(0, 6):
                    if k<3:
                        assert a[i,j,k] == stress[i,j].C[k,k]
                    elif k==3:
                        assert a[i,j,k] == stress[i,j].C[1,2]
                    elif k==4:
                        assert a[i,j,k] == stress[i,j].C[0,2]
                    else:
                        assert a[i,j,k] == stress[i,j].C[0,1]

    def test_tensile_stress(self):
        """Check that a stress tensor can be defined for tensile state"""
        n = 10
        sigma_11 = np.linspace(0,1, n)
        stress = Tensors.StressTensor.tensile([1,0,0], sigma_11)
        for i in range(0, n):
            stress_i = np.diag([sigma_11[i], 0, 0])
            np.testing.assert_array_equal(stress[i].matrix, stress_i)

    def test_shear_stress(self):
        """Check that a stress tensor can be defined for shear state"""
        n = 10
        sigma_12 = np.linspace(0,1, n)
        stress = Tensors.StressTensor.shear([1,0,0], [0,1,0], sigma_12)
        for i in range(0, n):
            stress_i = np.zeros((3,3))
            stress_i[0,1] = stress_i[1,0] = sigma_12[i]
            np.testing.assert_array_equal(stress[i].matrix, stress_i)

        # Now check if error is thrown if the two vectors are not orthogonal
        with self.assertRaises(ValueError) as context:
            Tensors.StrainTensor.shear([1,0,0], [1,1,0], 0.1)
        self.assertEqual(str(context.exception), 'u and v must be orthogonal')

    def test_set_item(self):
        """Check setting a tensor in a tensor array"""
        stress = Tensors.StressTensor.zeros((3, 3))
        stress[0,0] = np.ones(3)
        matrix = np.zeros((3, 3, 3, 3))
        matrix[0, 0, :, :] = 1
        np.testing.assert_array_equal(stress.matrix, matrix)

    def test_add_sub_mult_strain(self):
        """Check addition, subtraction and float multiplication of tensors"""
        shape = (3,3,3)
        a = Tensors.StrainTensor.ones(shape)
        b = 2 * Tensors.StrainTensor.ones(shape)
        c = 3 * Tensors.StrainTensor.ones(shape)
        d = a + b - c + 5 - 5 - a
        np.testing.assert_array_equal(d.matrix, -np.ones(shape + (3,3)))

    def test_flatten(self):
        """Check flattening of a tensor array"""
        shape = (3,3,3)
        matrix = np.random.random(shape + (3,3))
        a = SecondOrderTensor(matrix)
        a_flat = a.flatten()

        # Fist, check that the shapes are consistent
        assert a_flat.shape == np.prod(shape)

        # Then, check out each element
        for p in range(0, np.prod(shape)):
            i, j, k = np.unravel_index(p, shape)
            np.testing.assert_array_equal(a_flat[p].matrix, a[i,j,k].matrix)

    def test_symmetric_skew_parts(self):
        """Check the values returned by the symmetric and skew parts of a tensor"""
        shape = (2,3,4)
        a = SecondOrderTensor(np.random.random(shape + (3,3)))
        a_symm = a.symmetric_part()
        a_skew = a.skew_part()
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    matrix = a[i,j,k].matrix
                    np.testing.assert_array_equal(2 * a_symm[i, j, k].matrix, matrix + matrix.T )
                    np.testing.assert_array_equal(2 * a_skew[i, j, k].matrix, matrix - matrix.T)

    def test_equality(self):
        """Test the == operator"""
        # Test equality for two tensors of the same shape
        shape = (3,4,5)
        a = SecondOrderTensor(np.random.random(shape + (3,3)))
        b = SecondOrderTensor(np.random.random(shape + (3,3)))
        a[0,1,2] = b[0,1,2]
        is_equal = a == b
        assert is_equal.shape == shape
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    assert is_equal[i ,j, k] == np.all(a[i, j, k,:,:].matrix == b[i, j, k,:,:].matrix, axis=(-2,-1))

        # Test equality for an array of tensors, and a single tensor
        c = a[2, 1, 0]
        is_equal = a == c
        assert is_equal.shape == shape
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    assert is_equal[i ,j, k] == np.all(a[i, j, k,:,:].matrix == c.matrix, axis=(-2,-1))

        # Now test inconsistent shapes
        shape2 = (3,4,5,6)
        d = SecondOrderTensor(np.random.random(shape2 + (3, 3)))
        expected_error = 'The value to compare must be an array of shape {} or {}'.format(shape, shape + (3,3))
        with self.assertRaises(ValueError) as context:
            _ = a == d
        self.assertEqual(str(context.exception), expected_error)

    def test_divergence(self):
        """Test the divergence operator"""
        spacing = [0.1, 0.2, 0.3]
        x = np.arange(0, 1, spacing[0])
        y = np.arange(0, 1, spacing[1])
        z = np.arange(0, 1, spacing[2])
        x, y, z = np.meshgrid(x, y, z, indexing='ij')
        shape = x.shape
        a_11, b_11, c_11 = 2., 3., -1.
        a_12, b_12, c_12 = 4, -1., 2.
        a_23, b_23, c_23 = 1, -1, 3.

        s = SecondOrderTensor.zeros(shape=shape)
        s.C[0, 0] = a_11 * x + b_11 * y + c_11 * z
        s.C[0, 1] = a_12 * x + b_12 * y + c_12 * z
        s.C[1, 2] = a_23 * x + b_23 * y + c_23 * z
        div_space = s.div(spacing=spacing)
        div_nonspaced = s.div()
        div_uniaxial = s.div(axes=0)
        div_biaxial = s.div(axes=(0,1))
        expected_div_spaced = [a_11 + b_12, c_23, 0]
        expected_div_nonspaced = [a_11 * spacing[0] + b_12 * spacing[1], c_23 * spacing[2], 0]
        expected_div_uniaxial = [a_11 * spacing[0], 0, 0]
        expected_div_biaxial = [a_11 * spacing[0] + b_12 * spacing[1], 0, 0]
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    np.testing.assert_almost_equal(div_space[i, j, k], expected_div_spaced)
                    np.testing.assert_almost_equal(div_nonspaced[i, j, k], expected_div_nonspaced)
                    np.testing.assert_almost_equal(div_uniaxial[i, j, k], expected_div_uniaxial)
                    np.testing.assert_almost_equal(div_biaxial[i, j, k], expected_div_biaxial)


    def test_reshape(self):
        """Test reshaping and 'un-reshaping' an array of tensors"""
        init_shape = (6, 4, 5)
        new_shape = (30, 4)
        a = np.random.random(init_shape + (3, 3))
        t = SecondOrderTensor(a)
        t_reshaped = t.reshape(new_shape)
        assert t_reshaped.shape == new_shape
        t_reshaped_back = t_reshaped.reshape(init_shape)
        for i in range(0, init_shape[0]):
            for j in range(0, init_shape[1]):
                for k in range(0, init_shape[2]):
                    old_mat = a[i,j,k]
                    new_mat = t_reshaped_back[i,j,k].matrix
                    np.testing.assert_array_equal(old_mat, new_mat)

    def test_save_load_tensor(self):
        """Test save and load data to/from file"""

        # First, check with consistent shape
        a = np.random.random((5, 4, 3, 3))
        t = SecondOrderTensor(a)
        file_name = 'test_save.npy'
        t.save(file_name)
        t2 = t.load_from_npy(file_name)
        np.testing.assert_array_equal(t.matrix, t2.matrix)

        # Now try with inconsistent shape
        np.save(file_name, a[...,0])
        expected_error = 'The shape of the array to load must be (...,3,3).'
        with self.assertRaises(ValueError) as context:
            _ = SecondOrderTensor.load_from_npy(file_name)
        self.assertEqual(str(context.exception), expected_error)

    def test_save_load_csv(self):
        """Test saving and reading a CSV file"""
        # Check that when exporting/importing, we get the same tensor
        a = np.random.random((10, 3, 3))
        t = SecondOrderTensor(a)
        file_name = 'test_textfile.txt'
        t.save_as_txt(file_name)
        t2 = SecondOrderTensor.load_from_txt(file_name)
        np.testing.assert_array_almost_equal(t.matrix, t2.matrix)

        # Try with non-flatten tensor
        a = np.random.random((5, 3, 3, 3))
        t = SecondOrderTensor(a)
        expected_error = 'The array must be flatten before getting dumped to text file.'
        with self.assertRaises(ValueError) as context:
            t.save_as_txt(file_name)
        self.assertEqual(str(context.exception), expected_error)

        # Now try with a symmetric tensor
        a = np.random.random((10, 3, 3))
        a = a + np.swapaxes(a, -1, -2)
        t = SymmetricSecondOrderTensor(a)
        file_name = 'test_symmetric_textfile.txt'
        t.save_as_txt(file_name)
        t2 = SymmetricSecondOrderTensor.load_from_txt(file_name)
        np.testing.assert_array_almost_equal(t.matrix, t2.matrix)


    def test_symmetric_tensor_constructor(self):
        """Test constructor for symmetric second Order tensors"""

        # When a symmetric matrix is passed to the constructor
        mat = np.random.random((3,3))
        sym_mat = mat + mat.T
        t = SymmetricSecondOrderTensor(sym_mat)
        np.testing.assert_array_equal(t.matrix, sym_mat)

        # When an upper-diagonal matrix is passed to the constructor
        upper_mat = sym_mat
        upper_mat[np.tril_indices(3, -1)] = 0 # Set lower part to zero
        t2 = SymmetricSecondOrderTensor(upper_mat)
        assert t == t2

        # Expect error in any other case
        expected_error = ('The input array must be either slices of symmetric matrices, of slices of upper-diagonal '
                          'matrices.')
        with self.assertRaises(ValueError) as context:
            _ = SymmetricSecondOrderTensor(mat)
        self.assertEqual(str(context.exception), expected_error)

    def test_skew_symmetric_tensor_constructor(self):
        """Test constructor for symmetric second Order tensors"""

        # When a symmetric matrix is passed to the constructor
        mat = np.random.random((3,3))
        skew_sym_mat = mat - mat.T
        t = SkewSymmetricSecondOrderTensor(skew_sym_mat)
        np.testing.assert_array_equal(t.matrix, skew_sym_mat)

        # When an upper-diagonal matrix is passed to the constructor
        upper_mat = skew_sym_mat
        upper_mat[np.tril_indices(3)] = 0 # Set lower part to zero
        t2 = SkewSymmetricSecondOrderTensor(upper_mat)
        assert t == t2

        # Expect error in any other case
        expected_error = ('The input array must be either slices of symmetric matrices, of slices of upper-diagonal '
                          'matrices.')
        with self.assertRaises(ValueError) as context:
            _ = SymmetricSecondOrderTensor(mat)
        self.assertEqual(str(context.exception), expected_error)

    def test_add_sub_skew_symmetric_tensor(self):
        """Test basic operations between symmetric and skew-symmetric tensor, and check that the output classes are
        consistent."""
        mat = np.random.random((3,3))
        strain = StrainTensor(mat + mat.T)
        spin = SkewSymmetricSecondOrderTensor(mat - mat.T)

        # Check with consistent add.sub
        assert isinstance(strain + 2 * strain, StrainTensor)
        assert isinstance(spin + 2 * spin, SkewSymmetricSecondOrderTensor)

        # Check with inconsistent classes
        a = strain + 2 * spin
        assert (isinstance(a, SecondOrderTensor) and not isinstance(a, SymmetricSecondOrderTensor) and
                not isinstance(a, SkewSymmetricSecondOrderTensor))

        # Check negative values
        assert isinstance(-strain, StrainTensor)
        assert isinstance(-spin, SkewSymmetricSecondOrderTensor)

        # Check when adding a scaler value
        b = spin + 5
        assert (isinstance(b, SecondOrderTensor) and not isinstance(b, SymmetricSecondOrderTensor) and
                not isinstance(b, SkewSymmetricSecondOrderTensor))
        assert isinstance(strain + 5, StrainTensor)

        # Check when multiplying the tensor
        c = strain * spin
        assert (isinstance(c, SecondOrderTensor) and not isinstance(c, SymmetricSecondOrderTensor) and
                not isinstance(c, SkewSymmetricSecondOrderTensor))

        # Now check with rotations
        rotations = Rotation.random(100)
        assert isinstance(strain * rotations, StrainTensor)
        assert isinstance(spin * rotations, SkewSymmetricSecondOrderTensor)

    def test_to_pymatgen(self):
        """Test convertion from Elasticipy to pymatgen"""
        # First, try with a single tensor
        a = np.random.random((3, 3))
        a_sym = a + a.T
        strain = StrainTensor(a_sym)
        Strain_pymatgen = strain.to_pymatgen()
        np.testing.assert_array_equal(Strain_pymatgen.__array__(), a_sym)
        assert isinstance(Strain_pymatgen, mgStrain)

        # Now try with a 1D array
        n = 10
        b = np.random.random((n, 3, 3))
        b_sym = b + np.swapaxes(b, -1, -2)
        strain = StrainTensor(b_sym)
        Strain_pymatgen = strain.to_pymatgen()
        for i in range(n):
            np.testing.assert_array_equal(Strain_pymatgen[i].__array__(), b_sym[i])

        # Finally, try with a multidimensional array
        c = np.random.random((n, n, 3, 3))
        c_sym = c + np.swapaxes(c, -1, -2)
        strain = StrainTensor(c_sym)
        expected_error = 'The array must be flattened (1D tensor array) before converting to pytmatgen.'
        with self.assertRaises(ValueError) as context:
            _ = strain.to_pymatgen()
        self.assertEqual(str(context.exception), expected_error)

        # Now check that this also works for stress
        stress = StressTensor(a_sym)
        stress_pymatgen = stress.to_pymatgen()
        assert isinstance(stress_pymatgen, mgStress)

    def test_rand(self):
        """Test uniform random generation"""
        # Test two ways to define a rand tensor
        shape = (5,4)
        seed = 1324 # Ensure reproducibility
        t1 = SecondOrderTensor.rand(shape=shape, seed=seed)
        rng = np.random.default_rng(seed)
        t2 = SecondOrderTensor(rng.random(shape + (3,3)))
        assert np.all(t1==t2)

    def test_randn(self):
        """Test normal random distribution"""
        shape = (50, 40, 30)
        mean = np.random.random((3,3))
        std = np.random.random((3,3))
        t = SecondOrderTensor.randn(mean=mean, std=std, shape=shape)
        tmean = t.mean()
        tstd = t.std()
        tol = 1e-5
        np.testing.assert_array_almost_equal(tmean.matrix, mean, decimal=tol)
        np.testing.assert_array_almost_equal(tstd.matrix, std, decimal=tol)

    def test_inv(self):
        """Test inverse method"""
        shape = (3,4)
        matrix = np.random.random(shape + (3,3))
        t = SecondOrderTensor(matrix)
        tinv = t.inv()
        for i in range(shape[0]):
            for j in range(shape[1]):
                np.testing.assert_array_almost_equal(tinv[i,j].matrix, np.linalg.inv(matrix[i,j]))

    def test_tensile_shear(self):
        # Start with single tensile
        mag = 5
        t = Tensors.SymmetricSecondOrderTensor.tensile([1, 0, 0], mag)
        assert t.shape == ()
        mat = np.zeros((3, 3))
        mat[0, 0] = mag
        np.testing.assert_almost_equal(t.matrix, mat)

        # Now with arrays
        n = 5
        t = Tensors.SymmetricSecondOrderTensor.tensile([1,0,0], range(n))
        assert t.shape == (n,)
        mat = np.zeros((n,3,3))
        mat[:,0,0] = range(n)
        np.testing.assert_almost_equal(t.matrix, mat)

        # Try with a 2d directions
        with self.assertRaises(ValueError) as context:
            _ = Tensors.SymmetricSecondOrderTensor.tensile([1,0], range(n))
        self.assertEqual(str(context.exception), 'u must be 3D vector.')

        # Now shear
        t = Tensors.SymmetricSecondOrderTensor.shear([1,0,0], [0,1,0], range(n))
        assert t.shape == (n,)
        mat = np.zeros((n,3,3))
        mat[:,0,1] = mat[:,1,0] = range(n)
        np.testing.assert_almost_equal(t.matrix, mat)

    def test_repr(self):
        a=Tensors.StrainTensor.ones()
        assert a.__repr__() == 'Strain tensor\n[[1. 1. 1.]\n [1. 1. 1.]\n [1. 1. 1.]]'

        n=5
        b=Tensors.StressTensor.ones(n)
        assert b.__repr__() == 'Stress tensor\nShape=({},)'.format(n)

if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from pytest import approx
import os
import pandas as pd

from Elasticipy.FourthOrderTensor import StiffnessTensor, ComplianceTensor
from scipy.spatial.transform import Rotation
from Elasticipy.FourthOrderTensor import _indices2str
from Elasticipy.CrystalSymmetries import SYMMETRIES
from Elasticipy.StressStrainTensors import StressTensor
from pymatgen.analysis.elasticity import elastic as mg
from orix.quaternion import Rotation as orix_rot

current_dir = os.path.dirname(os.path.abspath(__file__))
file_path = os.path.join(current_dir, 'MaterialsProject.json')
data_base = pd.read_json(file_path)
rotations = Rotation.random(10000)

def variant_selection(symmetry, variant_name):
    for variant_group in symmetry.keys():
        elements = [elem.strip() for elem in variant_group.split(",")]
        if variant_name in elements:
            return symmetry[variant_group]
    return None

Smat = np.array([[8, -3, -2, 0, 14, 0],
                 [-3, 8, -5, 0, -8, 0],
                 [-2, -5, 10, 0, 0, 0],
                 [0, 0, 0, 12, 0, 0],
                 [14, -8, 0, 0, 116, 0],
                 [0, 0, 0, 0, 0, 12]])/1000
S = ComplianceTensor(Smat)


def crystal_symmetry_tester(symmetry_name, cls='stiffness', variant=None):
    symmetry = SYMMETRIES[symmetry_name]
    if variant is None:
        materials_of_interest = data_base[data_base.symmetry == symmetry_name]
        required_fields = symmetry.required
    else:
        materials_of_interest = data_base[data_base.point_group == variant]
        variant = variant_selection(symmetry, variant)
        required_fields = variant.required
    for index, row in materials_of_interest.iterrows():
        matrix = np.array(row['C'])
        if cls=='stiffness':
            class_constructor = StiffnessTensor
        else:
            class_constructor = ComplianceTensor
            matrix = np.linalg.inv(matrix)*1000
        kwargs = dict()
        for indices in required_fields:
            component_name = 'C' + _indices2str(indices)
            kwargs[component_name] = matrix[tuple(indices)]
        constructor = getattr(class_constructor, symmetry_name.lower())
        C = constructor(**kwargs)
        assert np.all(C.matrix == approx(matrix, rel=0.5))


class TestComplianceTensor(unittest.TestCase):
    def test_young_modulus_eval(self):
        E = S.Young_modulus
        E_xyz = E.eval(np.eye(3))
        for i in range(3):
            self.assertEqual(E_xyz[i], 1/Smat[i, i])

    def test_young_modulus_stats(self):
        E = S.Young_modulus
        assert E.mean(method='exact') == approx(101.994)
        assert E.std() == approx(48.48264174566468)
        assert E.mean(method='trapezoid') == approx(101.9855, rel=1e-3)

    def test_shear_modulus_eval(self):
        G = S.shear_modulus
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        G_xyz = G.eval(u, v)
        for i in range(3):
            self.assertEqual(G_xyz[i],  1/Smat[i+3, i+3])

    def test_Poisson_ratio_eval(self):
        nu = S.Poisson_ratio
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        nu_xyz = nu.eval(u, v)
        nu_xyz_th = [0.625, 0.25, 0.375]
        for i in range(3):
            self.assertEqual(nu_xyz[i],  nu_xyz_th[i])

    def test_shear_modulus_mini_maxi(self):
        G = S.shear_modulus
        G_min, _ = G.min()
        G_max, _ = G.max()
        assert G_min == approx(8.47165)
        assert G_max == approx(83.3333)

    def test_unvoigt(self):
        lame1, lame2 = 1, 2
        C = StiffnessTensor.fromCrystalSymmetry(C11=lame1 + 2 * lame2,
                                                C12=lame1, symmetry='isotropic')
        C_full = C.full_tensor()
        eye = np.eye(3)
        A = np.einsum('ij,kl->ijkl', eye, eye)
        B = np.einsum('ik,jl->ijkl', eye, eye)
        C = np.einsum('il,kj->ijkl', eye, eye)
        C_th = lame1 * A + lame2 * (B + C)
        np.testing.assert_almost_equal(C_th, C_full)

    def test_averages(self):
        averages = [S.Voigt_average(), S.Reuss_average(), S.Hill_average()]
        E_mean_th = [151.738, 75.76, 114.45]
        G_mean_th = [55.653, 26.596, 41.124]
        nu_mean_th = [0.36325, 0.42424242, 0.3915]
        for i, average in enumerate(averages):
            assert approx(average.Young_modulus.eval([1,0,0]), rel=1e-4) == E_mean_th[i]
            assert approx(average.shear_modulus.eval([1,0,0],[0,1,0]), rel=1e-4) == G_mean_th[i]
            assert approx(average.Poisson_ratio.eval([1,0,0],[0,1,0]), rel=1e-4) == nu_mean_th[i]

    def test_isotropic(self, E=210000, nu=0.28):
        C = StiffnessTensor.isotropic(E=E, nu=nu)
        G = C.shear_modulus.mean()
        assert approx(G) == E / (1+nu) /2
        C = StiffnessTensor.isotropic(E=E, lame2=G)
        assert approx(C.Poisson_ratio.mean()) == nu
        C = StiffnessTensor.isotropic(lame2=G, nu=nu)
        assert approx(C.Young_modulus.mean()) == E

    def test_wave_velocity(self, E=210, nu=0.3, rho=7.8):
        C = StiffnessTensor.isotropic(E=E, nu=nu)
        M = E * (1 - nu) / ((1 + nu) * (1 - 2 * nu))
        cp, cs_1, cs_2 = C.wave_velocity(rho)
        assert approx(cp.mean()) == np.sqrt(M / rho)
        G = C.shear_modulus.mean()
        assert approx(cs_2.mean()) == np.sqrt(G / rho)
        assert approx(cs_1.mean()) == np.sqrt(G / rho)

    def test_symmetry(self):
        S = np.random.random((6, 6))
        with self.assertRaises(ValueError) as context:
            _ = ComplianceTensor(S)
        self.assertEqual(str(context.exception), 'The input matrix must be symmetric')

    def test_positive_definite(self):
        S = np.array([
            [2, -1, 0, 0, 0, 0],
            [-1, 2, -1, 0, 0, 0],
            [0, -1, 2, -1, 0, 0],
            [0, 0, -1, 2, -1, 0],
            [0, 0, 0, -1, 2, -1],
            [0, 0, 0, 0, -1, 0]
        ])
        with self.assertRaises(ValueError) as context:
            _ = ComplianceTensor(S)
        eig_vals = np.linalg.eigvals(S)
        expected_error = 'The input matrix is not definite positive (eigenvalues: {})'.format(eig_vals)
        self.assertEqual(str(context.exception), expected_error)

    def test_input_matrix(self):
        for n in range(1,8):
            m = np.ones(shape=(n,n))
            if n != 6:
                with self.assertRaises(ValueError) as context:
                    _ = ComplianceTensor(m)
                self.assertEqual(str(context.exception), 'The input matrix must of shape (6,6)')

    def test_full_tensor_as_input(self):
        a = ComplianceTensor.isotropic(E=210, nu=0.3)
        b = ComplianceTensor(a.full_tensor())
        np.testing.assert_array_almost_equal(a.matrix, b.matrix)


    def test_component(self):
        assert S.C11 == Smat[0, 0]
        assert S.C12 == Smat[0, 1]
        assert S.C13 == Smat[0, 2]
        assert S.C14 == Smat[0, 3]
        assert S.C15 == Smat[0, 4]
        assert S.C16 == Smat[0, 5]
        assert S.C22 == Smat[1, 1]
        assert S.C23 == Smat[1, 2]
        assert S.C24 == Smat[1, 3]
        assert S.C25 == Smat[1, 4]
        assert S.C26 == Smat[1, 5]
        assert S.C32 == Smat[2, 1]
        assert S.C33 == Smat[2, 2]
        assert S.C34 == Smat[2, 3]
        assert S.C35 == Smat[2, 4]
        assert S.C36 == Smat[2, 5]
        assert S.C44 == Smat[3, 3]
        assert S.C45 == Smat[3, 4]
        assert S.C46 == Smat[3, 5]
        assert S.C55 == Smat[4, 4]
        assert S.C56 == Smat[4, 5]
        assert S.C66 == Smat[5, 5]
        docstring = ComplianceTensor.C12.__doc__
        assert docstring == 'Returns the (1,2) component of the Compliance matrix.'

    def test_bulk_modulus(self):
        E, nu = 210, 0.3
        Siso= ComplianceTensor.isotropic(E=E, nu=nu)
        assert Siso.bulk_modulus == approx(E / (3 * (1-2 * nu)))

class TestStiffnessConstructor(unittest.TestCase):
    def test_averages(self):
        """Check that the Voigt, Reuss and Hill averages are consistent with those provided by MP."""
        rel = 5e-2
        for index, row in data_base.iterrows():
            matrix = row['C']
            symmetry = row['symmetry']
            C = StiffnessTensor(matrix, symmetry=symmetry)
            C_rotated = C * rotations
            for method in ('voigt', 'reuss', 'hill', 'dummy'):
                if method == 'dummy':
                    with self.assertRaises(NotImplementedError) as context:
                        _ = C_rotated.average(method)
                    self.assertEqual(str(context.exception), 'Only Voigt, Reus, and Hill are implemented.')
                else:
                    Gavg = C.average(method).shear_modulus.mean(n_evals=10000)
                    assert row['G' + method] == approx(Gavg, rel=rel)
                    Gavg = C_rotated.average(method).shear_modulus.mean(n_evals=10000)
                    assert row['G' + method] == approx(Gavg, rel=rel)

    def test_stiffness_cubic(self):
        """Check that all symmetries in stiffness are well taken into account for cubic case"""
        crystal_symmetry_tester('Cubic')

    def test_stiffness_hexagonal(self):
        """Check that all symmetries in stiffness are well taken into account for hexagonal case"""
        crystal_symmetry_tester('Hexagonal')

    def test_stiffness_trigonal(self):
        """Check that all symmetries in stiffness are well taken into account for trigonal case"""
        crystal_symmetry_tester('Trigonal', variant='32')
        crystal_symmetry_tester('Trigonal', variant='-3')

    def test_stiffness_tetragonal(self):
        """Check that all symmetries in stiffness are well taken into account for tetragonal case"""
        crystal_symmetry_tester('Tetragonal', variant='-42m')
        crystal_symmetry_tester('Tetragonal', variant='-4')

    def test_stiffness_orthorhombic(self):
        """Check that all symmetries in stiffness are well taken into account for orthorhombic case"""
        crystal_symmetry_tester('Orthorhombic')

    def test_stiffness_monoclinic(self):
        """Check that all symmetries in stiffness are well taken into account for monoclinic case"""
        crystal_symmetry_tester('Monoclinic', variant='Diad || y')

    def test_compliance_cubic(self):
        """Check that all symmetries in compliance are well taken into account for cubic case"""
        crystal_symmetry_tester('Cubic', cls='compliance')

    def test_compliance_hexagonal(self):
        """Check that all symmetries in compliance are well taken into account for hexagonal case"""
        crystal_symmetry_tester('Hexagonal', cls='compliance')

    def test_compliance_trigonal(self):
        """Check that all symmetries in compliance are well taken into account for trigonal case"""
        crystal_symmetry_tester('Trigonal', variant='32', cls='compliance')
        crystal_symmetry_tester('Trigonal', variant='-3', cls='compliance')

    def test_compliance_tetragonal(self):
        """Check that all symmetries in compliance are well taken into account for tetragonal case"""
        crystal_symmetry_tester('Tetragonal', variant='-42m', cls='compliance')
        crystal_symmetry_tester('Tetragonal', variant='-4', cls='compliance')

    def test_compliance_orthorhombic(self):
        """Check that all symmetries in compliance are well taken into account for orthorhombic case"""
        crystal_symmetry_tester('Orthorhombic', cls='compliance')

    def test_compliance_monoclinic(self):
        """Check that all symmetries in compliance are well taken into account for monoclinic case"""
        crystal_symmetry_tester('Monoclinic', variant='Diad || y', cls='compliance')

    def test_young_modulus_eval(self):
        """Check that the Young modulus is given somehow given by the compliance tensor"""
        E = S.Young_modulus
        E_xyz = E.eval(np.eye(3))
        for i in range(3):
            self.assertEqual(E_xyz[i], 1/Smat[i, i])

    def test_young_modulus_stats(self):
        """Test statistics for Young moduli"""
        E = S.Young_modulus
        assert E.mean(method='exact') == approx(101.994)
        assert E.std() == approx(48.48264174566468)
        assert E.mean() == approx(101.9942123)

    def test_shear_modulus_eval(self):
        """Test shear moduli estimations"""
        G = S.shear_modulus
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        G_xyz = G.eval(u, v)
        for i in range(3):
            self.assertEqual(G_xyz[i],  1/Smat[i+3, i+3])

    def test_Poisson_ratio_eval(self):
        """Test Poisson ration estimations"""
        nu = S.Poisson_ratio
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        nu_xyz = nu.eval(u, v)
        nu_xyz_th = [0.625, 0.25, 0.375]
        for i in range(3):
            self.assertEqual(nu_xyz[i],  nu_xyz_th[i])

    def test_shear_modulus_mini_maxi(self):
        """Test shear min/max"""
        G = S.shear_modulus
        G_min, _ = G.min()
        G_max, _ = G.max()
        assert G_min == approx(8.47165)
        assert G_max == approx(83.3333)

    def test_unvoigt(self):
        """Test if the isotropic second-order tensor is well reconstructed"""
        lame1, lame2 = 1, 2
        C = StiffnessTensor.fromCrystalSymmetry(C11=lame1 + 2 * lame2,
                                                C12=lame1, symmetry='isotropic')
        C_full = C.full_tensor()
        eye = np.eye(3)
        A = np.einsum('ij,kl->ijkl', eye, eye)
        B = np.einsum('ik,jl->ijkl', eye, eye)
        C = np.einsum('il,kj->ijkl', eye, eye)
        C_th = lame1 * A + lame2 * (B + C)
        np.testing.assert_almost_equal(C_th, C_full)

    def test_isotropic(self):
        """Test if we correctly retrieve engineering constants in the isotropic case"""
        E = 210000
        nu = 0.28
        C = StiffnessTensor.isotropic(E=E, nu=nu)
        G = C.shear_modulus.mean()
        assert approx(G) == E / (1+nu) /2
        C = StiffnessTensor.isotropic(E=E, lame2=G)
        assert approx(C.Poisson_ratio.mean()) == nu
        C = StiffnessTensor.isotropic(lame2=G, nu=nu)
        assert approx(C.Young_modulus.mean()) == E

    def test_wave_velocity(self):
        """Test computation of wave velocities against simple isotropic case"""
        E = 210
        nu = 0.3
        rho = 7.8
        C = StiffnessTensor.isotropic(E=E, nu=nu)
        M = E * (1 - nu) / ((1 + nu) * (1 - 2 * nu))
        cp, cs_1, cs_2 = C.wave_velocity(rho)
        assert approx(cp.mean()) == np.sqrt(M / rho)
        G = C.shear_modulus.mean()
        assert approx(cs_2.mean()) == np.sqrt(G / rho)
        assert approx(cs_1.mean()) == np.sqrt(G / rho)

    def test_len(self):
        """Test len(ght) operator"""
        C = StiffnessTensor.isotropic(E=210000, nu=0.3)
        assert len(C) == 1
        assert len(C * rotations[0]) == 1
        assert len(C * rotations) == 10000

    def test_monoclinic(self):
        """Test constructor for monoclinic symmetry"""
        common_arguments = {'C11':11, 'C12':12, 'C13':13, 'C22':22, 'C23':23, 'C33':33, 'C44':44, 'C55':55, 'C66':66}

        # Check for Diad||y
        C = StiffnessTensor.monoclinic(**common_arguments, C16=16, C26=26, C36=36, C45=45)
        matrix = np.array([[11, 12, 13, 0, 0, 16],
                           [12, 22, 23, 0, 0, 26],
                           [13, 23, 33, 0, 0, 36],
                           [0,  0,  0, 44, 45, 0],
                           [0,  0,  0, 45, 55, 0],
                           [16, 26, 36, 0, 0, 66]], dtype=np.float64)
        np.testing.assert_array_equal(matrix, C.matrix)

        # Check for Diad||z
        C = StiffnessTensor.monoclinic(**common_arguments, C15=15, C25=25, C35=35, C46=46)
        matrix = np.array([[11, 12, 13, 0, 15, 0],
                           [12, 22, 23, 0,  25, 0],
                           [13, 23, 33, 0,  35, 0],
                           [0,  0,  0,  44, 0, 46],
                           [15, 25, 35, 0,  55, 0],
                           [0,  0,  0,  46, 0, 66]], dtype=np.float64)
        np.testing.assert_array_equal(matrix, C.matrix)

        # Check ambiguous cases
        expected_error = "'Ambiguous diad. Provide either C15, C25, C35 and C46; or C16, C26, C36 and C45'"
        with self.assertRaises(KeyError) as context:
            C = StiffnessTensor.monoclinic(**common_arguments,
                                          C15=15, C25=25, C35=35, C46=46, C16=16, C26=26, C36=36, C45=45)
        self.assertEqual(str(context.exception), expected_error)

        expected_error = ("'For monoclinic symmetry, one should provide either C15, C25, C35 and C46, "
                          "or C16, C26, C36 and C45.'")
        with self.assertRaises(KeyError) as context:
            C = StiffnessTensor.monoclinic(**common_arguments)
        self.assertEqual(str(context.exception), expected_error)

    def test_write_read_tensor(self):
        """Test export and import stiffness tensor to text file"""
        C = StiffnessTensor.isotropic(E=210, nu=0.3)
        filename = 'C_tmp.txt'
        C.save_to_txt(filename)
        C2 = StiffnessTensor.from_txt_file(filename)
        np.testing.assert_allclose(C2.matrix, C.matrix, atol=1e-2)

    def test_equality(self):
        """Test == operator"""
        C1 = StiffnessTensor.isotropic(E=210000, nu=0.3)
        C2 = StiffnessTensor.isotropic(E=210000, nu=0.3)
        assert C1 == C2
        assert C1 == C2.matrix

    def test_add_sub(self):
        """Test addition and subtraction of tensors"""
        C1 = StiffnessTensor.isotropic(E=200, nu=0.3)
        C2 = StiffnessTensor.isotropic(E=100, nu=0.3)
        C_plus = C1 + C2
        assert C_plus.Young_modulus.mean() == approx(300)
        C_minus = C1 - C2
        assert C_minus.Young_modulus.mean() == approx(100)
        C_minus = C1 - C2.matrix
        assert C_minus.Young_modulus.mean() == approx(100)
        C_plus_full = C1 + C2.full_tensor()
        assert C_plus_full == C_plus

        with self.assertRaises(ValueError) as context:
            _ = C1 + C2.inv()
        self.assertEqual(str(context.exception), 'The two tensors to add must be of the same class.')

    def test_mul_rmul(self):
        C = StiffnessTensor.isotropic(E=200, nu=0.3)
        C1 = C * 2
        C2 = 2 * C
        assert C1 == C2

    def test_div(self):
        C = StiffnessTensor.isotropic(E=200, nu=0.3)
        Cdiv = C/2
        np.testing.assert_array_equal(Cdiv.matrix, C.matrix/2)

    def test_repr(self):
        """Test printing out the tensor"""
        C = StiffnessTensor.isotropic(E=210000, nu=0.3)
        str = C.__repr__()
        assert str == ('Stiffness tensor (in Voigt notation):\n'
                       '[[282692.30769231 121153.84615385 121153.84615385      0.\n'
                       '       0.              0.        ]\n'
                       ' [121153.84615385 282692.30769231 121153.84615385      0.\n'
                       '       0.              0.        ]\n'
                       ' [121153.84615385 121153.84615385 282692.30769231      0.\n'
                       '       0.              0.        ]\n'
                       ' [     0.              0.              0.          80769.23076923\n'
                       '       0.              0.        ]\n'
                       ' [     0.              0.              0.              0.\n'
                       '   80769.23076923      0.        ]\n'
                       ' [     0.              0.              0.              0.\n'
                       '       0.          80769.23076923]]\nSymmetry: isotropic')

        C = StiffnessTensor.cubic(C11=200, C12=100, C44=400, phase_name='Cu')
        str_Cu = ('Stiffness tensor (in Voigt notation) for Cu:\n'
                       '[[200. 100. 100.   0.   0.   0.]\n'
                       ' [100. 200. 100.   0.   0.   0.]\n'
                       ' [100. 100. 200.   0.   0.   0.]\n'
                       ' [  0.   0.   0. 400.   0.   0.]\n'
                       ' [  0.   0.   0.   0. 400.   0.]\n'
                       ' [  0.   0.   0.   0.   0. 400.]]\n'
                       'Symmetry: cubic')
        assert C.__repr__() == str_Cu

        C_rotated = C * rotations
        assert C_rotated.__repr__() == str_Cu + '\n{} orientations'.format(len(rotations))



    def test_weighted_average(self):
        """Test averaging two phases"""
        E1 = 100
        E2 = 200
        C1 = StiffnessTensor.isotropic(E=E1, nu=0.3)
        C2 = StiffnessTensor.isotropic(E=E2, nu=0.3)
        Cv = StiffnessTensor.weighted_average((C1, C2), [0.5, 0.5], method='Voigt')
        Cr = StiffnessTensor.weighted_average((C1, C2), [0.5, 0.5], method='Reuss')
        Ch = StiffnessTensor.weighted_average((C1, C2), [0.5, 0.5], method='Hill')
        E_voigt = (E1 + E2) / 2
        E_reuss = 2 / (1/E1 + 1/E2)
        assert Cv.Young_modulus.mean() == approx(E_voigt)
        assert Cr.Young_modulus.mean() == approx(E_reuss)
        assert Ch.Young_modulus.mean() == approx(E_voigt/2 + E_reuss/2)

    def test_orthotropic(self):
        """Check if the engineering constants are well retrieved in the orthotropic case"""
        Ex, Ey, Ez = 100., 200., 300.
        nu_yx, nu_zy, nu_zx = 0.2, 0.3, 0.4
        G_xy, G_xz, G_yz = 50., 60., 70.
        C = StiffnessTensor.orthotropic(Ex=Ex, Ey=Ey, Ez=Ez, nu_yx=nu_yx, nu_zx=nu_zx, nu_zy=nu_zy,
                                        Gxy=G_xy, Gxz=G_xz, Gyz=G_yz)
        E = C.Young_modulus
        assert E.eval([1,0,0]) == approx(Ex)
        assert E.eval([0,1,0]) == approx(Ey)
        assert E.eval([0,0,1]) == approx(Ez)
        G = C.shear_modulus
        assert G.eval([1,0,0], [0,1,0]) == approx(G_xy)
        assert G.eval([1, 0, 0], [0, 0, 1]) == approx(G_xz)
        assert G.eval([0, 1, 0], [0, 0, 1]) == approx(G_yz)
        nu = C.Poisson_ratio
        assert nu.eval([0,1,0], [1,0,0]) == approx(nu_yx)
        assert nu.eval([0, 0, 1], [0, 1, 0]) == approx(nu_zy)
        assert nu.eval([0, 0, 1], [1, 0, 0]) == approx(nu_zx)

    def test_transverse_isotropic(self):
        """Check if the engineering constants are well retrieved in the transverse-isotropic case"""
        Ex, Ez = 100., 200.
        nu_yx, nu_zx = 0.2, 0.3
        Gxz = 80
        C = StiffnessTensor.transverse_isotropic(Ex=Ex, Ez=Ez, nu_yx=nu_yx, nu_zx=nu_zx, Gxz=Gxz)
        E = C.Young_modulus
        assert E.eval([1,0,0]) == approx(Ex)
        assert E.eval([0,1,0]) == approx(Ex)
        assert E.eval([0,0,1]) == approx(Ez)
        G = C.shear_modulus
        assert G.eval([1, 0, 0], [0, 0, 1]) == approx(Gxz)
        nu = C.Poisson_ratio
        assert nu.eval([0,1,0], [1,0,0]) == approx(nu_yx)
        assert nu.eval([0, 0, 1], [0, 1, 0]) == approx(nu_zx)
        assert nu.eval([0, 0, 1], [1, 0, 0]) == approx(nu_zx)

    def test_straining_energy(self):
        """Test if the elastic energies are consistent."""
        matrix = np.random.random((3,3))
        stress = StressTensor(matrix + matrix.T)
        strain = S * stress
        e1 = stress.elastic_energy(strain)
        e2 = strain.elastic_energy(stress)
        np.testing.assert_approx_equal(e1, e2)


    def test_to_pymatgen(self):
        """Test exporting stiffness and compliance to pymatgen format"""
        C = S.inv()
        Cvrh = C.Hill_average()
        Cvrh_pymatgen = mg.ElasticTensor(Cvrh.full_tensor())
        C_pymatgen = mg.ElasticTensor(C.full_tensor())
        assert Cvrh_pymatgen.y_mod == approx(Cvrh.Young_modulus.mean()*1e9)
        assert C_pymatgen.g_vrh == approx(Cvrh.shear_modulus.mean())

        S_pymatgen = S.to_pymatgen()
        np.testing.assert_array_almost_equal(S_pymatgen.voigt, np.linalg.inv(C_pymatgen.voigt))

    def test_MaterialsProject(self):
        """Test import from the Materials Project"""
        # Try with cubic Cu
        C = StiffnessTensor.from_MP("mp-30")
        C_Cu = StiffnessTensor.cubic(C11=186, C12=134, C44=77)
        np.testing.assert_array_almost_equal(C.matrix, C_Cu.matrix)

        # Now try with a list of entries
        Cs = StiffnessTensor.from_MP(("mp-30", "mp-1048"))
        assert len(Cs) == 2
        np.testing.assert_array_almost_equal(Cs[0].matrix, C_Cu.matrix)


    def test_getitem(self):
        """Test indexing of stiffness tensor"""
        S_rotated = S * rotations
        S1 = S_rotated[0]
        S2 = S * rotations[0]
        np.testing.assert_array_almost_equal(S1.matrix, S2.matrix)
        expected_error = 'The tensor has no orientation, therefore it cannot be indexed.'
        with self.assertRaises(IndexError) as context:
            _ = S[0]
        self.assertEqual(str(context.exception), expected_error)

    def test_symmetry(self):
        S = np.random.random((6, 6))
        with self.assertRaises(ValueError) as context:
            _ = StiffnessTensor(S)
        self.assertEqual(str(context.exception), 'The input matrix must be symmetric')

    def test_positive_definite(self):
        S = np.array([
            [2, -1, 0, 0, 0, 0],
            [-1, 2, -1, 0, 0, 0],
            [0, -1, 2, -1, 0, 0],
            [0, 0, -1, 2, -1, 0],
            [0, 0, 0, -1, 2, -1],
            [0, 0, 0, 0, -1, 0]
        ])
        with self.assertRaises(ValueError) as context:
            _ = StiffnessTensor(S)
        eig_vals = np.linalg.eigvals(S)
        expected_error = 'The input matrix is not definite positive (eigenvalues: {})'.format(eig_vals)
        self.assertEqual(str(context.exception), expected_error)

    def test_Zener_universal_anisotropy(self):
        C11, C12, C44 = 173, 33, 18
        C = StiffnessTensor.cubic(C11=C11, C12=C12, C44=C44)
        Z = C.Zener_ratio
        assert 6/5 * (Z**0.5 - Z**(-0.5))**2 == approx(C.universal_anisotropy)
        C_cub_iso = StiffnessTensor.cubic(C11=C11, C12=C12, C44=(C11-C12)/2)
        assert C_cub_iso.Zener_ratio == 1.0
        Ciso = StiffnessTensor.isotropic(E=210, nu=0.3)
        assert Ciso.Zener_ratio == 1.0
        assert Ciso.universal_anisotropy == 0.0
        Cmono = S.inv()
        assert np.isnan(Cmono.Zener_ratio)

    def test_orix(self):
        n=5
        orix_rotations = orix_rot.random(n)
        C = S.inv()
        C_rotated = C * orix_rotations
        C_rotated_full = C_rotated.full_tensor()
        for i in range(n):
            rot_mat = orix_rotations.to_matrix()[i]
            tensor_i = np.einsum('im,jn,ko,lp,mnop -> ijkl', rot_mat, rot_mat, rot_mat, rot_mat, C.full_tensor())
            np.testing.assert_array_almost_equal(C_rotated_full[i], tensor_i)


    def test_linear_compressibility(self):
        E, nu = 210, 0.3
        Ciso = StiffnessTensor.isotropic(E=E, nu=nu)
        beta = Ciso.linear_compressibility.mean()
        assert Ciso.bulk_modulus == approx(1 / (3*beta))

    def test_full_tensor_as_input(self):
        a = StiffnessTensor.isotropic(E=210, nu=0.3)
        b = StiffnessTensor(a.full_tensor())
        assert a == b

    def test_component(self):
        C = S.inv()
        Cmat = C.matrix
        assert C.C11 == Cmat[0, 0]
        assert C.C12 == Cmat[0, 1]
        assert C.C13 == Cmat[0, 2]
        assert C.C14 == Cmat[0, 3]
        assert C.C15 == Cmat[0, 4]
        assert C.C16 == Cmat[0, 5]
        assert C.C22 == Cmat[1, 1]
        assert C.C23 == Cmat[1, 2]
        assert C.C24 == Cmat[1, 3]
        assert C.C25 == Cmat[1, 4]
        assert C.C26 == Cmat[1, 5]
        assert C.C32 == Cmat[2, 1]
        assert C.C33 == Cmat[2, 2]
        assert C.C34 == Cmat[2, 3]
        assert C.C35 == Cmat[2, 4]
        assert C.C36 == Cmat[2, 5]
        assert C.C44 == Cmat[3, 3]
        assert C.C45 == Cmat[3, 4]
        assert C.C46 == Cmat[3, 5]
        assert C.C55 == Cmat[4, 4]
        assert C.C56 == Cmat[4, 5]
        assert C.C66 == Cmat[5, 5]
        docstring = StiffnessTensor.C12.__doc__
        assert docstring == 'Returns the (1,2) component of the Stiffness matrix.'

    def test_bulk_modulus(self):
        E, nu = 210, 0.3
        Ciso= StiffnessTensor.isotropic(E=E, nu=nu)
        assert Ciso.bulk_modulus == approx(E / (3 * (1-2 * nu)))

if __name__ == '__main__':
    unittest.main()
